
#%%
class EZTract7DT:
    
    def __init__(self):
        pass
    


#%%


# %%
if __name__ == '__main__':
    from bridge.utils import HostGalaxyCatalog
    hg = HostGalaxyCatalog()
#%%
if __name__ == '__main__':
    from ezphot.utils import DataBrowser
    dbrowser = DataBrowser('scidata')
    dbrowser.objname = 'AT2025hp'
    target_imgset = dbrowser.search(pattern = 'coadd**.fits', return_type = 'science')
    target_imgset.select_images(filter = ['m400','m425','m450','m475','m500','m525','m550','m575','m600','m625','m650','m675','m700','m725','m750','m775','m800','m825','m850','m875'])
    target_imgset.target_images.sort(key = lambda x: x.filter)
    image_paths = [img.path for img in target_imgset.target_images]
    catalog_paths = [img.catalog.path for img in target_imgset.target_images]
    filter_list = [img.filter for img in target_imgset.target_images]
#%%
# ================================================
# ZP CALIBRATION
# ================================================
#%% FIND REFERENCE CATALOG
if __name__ == '__main__':
    target_refcatset = dbrowser.search(pattern = 'coadd*.fits.refcat', return_type = 'catalog')
    for refcat in target_refcatset.catalogs:
        refcat.data
        print(refcat.nsources)
    merged_tbl, merged_metadata = target_refcatset.merge_catalogs()       
# %% RUN TRACT7DT
if __name__ == '__main__':
    target_ra = [30.7663010688] # SN2025afih
    target_dec = [4.24367725701]
    target_ra = [101.796646]
    target_dec = [-10.420567]
    target_type = ['STAR']
    host_tbl = hg.search_catalog(ra = target_ra[0], dec = target_dec[0], radius_deg = 10/3600)
    if len(host_tbl) > 0:
        host_ra = host_tbl['RA'][0]
        host_dec = host_tbl['Dec'][0]
        host_type = 'EXP'
        target_ra.append(host_ra)
        target_dec.append(host_dec)
        target_type.append(host_type)        
#%%
if __name__ == '__main__':
    import numpy as np
    self = Tract7DTRunner(image_paths = image_paths, filter_list = filter_list, catalog_paths = catalog_paths)
    list_ra = list(merged_tbl['ra_basis']) 
    list_dec = list(merged_tbl['dec_basis']) 
    list_type = ['STAR'] * len(list_ra)
    list_ra.extend(target_ra)
    list_dec.extend(target_dec)
    list_type.extend(target_type)
    self.register_target(list_ra = list_ra, list_dec = list_dec, list_type = list_type, update_type_from_catalog = False, update_flux_from_catalog = False)
    self.register_reference_catalog(objname = dbrowser.objname)
    self.run()
# %% READ TRACT7DT RESULT
if __name__ == '__main__':
    from astropy.io import ascii
    result = ascii.read(self.workdir / 'final_catalog_with_fit.csv')
    flux_keys = [key for key in result.colnames if key.startswith('FLUX_') and not key.endswith('_fit')]
    filter_keys = []
    for flux_key in flux_keys:
        mag_key = flux_key.replace('FLUX_', 'MAG_')
        flux_key_fit = flux_key + '_fit'
        mag_key_fit = mag_key + '_fit'
        filter = flux_key.replace('FLUX_', '')
        filter_keys.append(filter)
        flux_input = result[flux_key]
        flux_fit = result[flux_key_fit]    
        mag_fit = -2.5*np.log10(flux_fit)
        result[mag_key_fit] = mag_fit
#%% MATCH PSF AND REF CATALOGS
if __name__ == '__main__':
    from ezphot.helper import Helper
    from astropy.coordinates import SkyCoord
    from ezphot.skycatalog import SkyCatalog
    refcatalog = SkyCatalog(objname = dbrowser.objname)
    ref_tbl = refcatalog.data

    max_distance_arcsec = 10
    helper = Helper()
    ra_psf = result['RA']
    dec_psf = result['DEC']
    skycoord_psf = SkyCoord(ra_psf, dec_psf, unit = 'deg')
    ra_ref = ref_tbl['ra']
    dec_ref = ref_tbl['dec']
    skycoord_ref = SkyCoord(ra_ref, dec_ref, unit = 'deg')
    idx_psf, idx_ref, _ = helper.cross_match(skycoord_psf, skycoord_ref, max_distance_second = max_distance_arcsec)
    result_matched = result[idx_psf]
    ref_tbl_matched = ref_tbl[idx_ref]
#%% CALCULATE ZP FOR EACH FILTER
if __name__ == '__main__':
    helper = Helper()
    zp_all = dict()
    mag_psf_all = dict()
    mag_ref_all = dict()
    for flux_key in flux_keys:
        mag_key_fit = flux_key.replace('FLUX_', 'MAG_') + '_fit'
        filter = flux_key.replace('FLUX_', '')
        mag_key_ref = f'{filter}_mag'
        mag_from_psf = result_matched[mag_key_fit]
        mag_from_ref = ref_tbl_matched[mag_key_ref]
        zp_all[filter] = mag_from_ref - mag_from_psf
        mag_psf_all[filter] = mag_from_psf
        mag_ref_all[filter] = mag_from_ref
#%% PLOT ZP FOR EACH FILTER
from astropy.stats import sigma_clipped_stats
if __name__ == '__main__':
    import matplotlib.pyplot as plt
    from ezphot.dataobjects import LightCurve
    zp_filter = dict()
    zp_filter_std = dict()
    color_map = LightCurve.FILTER_COLOR
    plt.figure(figsize = (10, 6))
    for filter_ in filter_list:
        mag_psf = mag_psf_all[filter_]
        mag_ref = mag_ref_all[filter_]
        zp = zp_all[filter_]
        
        zp_mean, zp_median, zp_std = sigma_clipped_stats(zp, sigma = 3, maxiters = 5)
        
        plt.scatter(mag_psf + zp_median, zp, edgecolor = color_map[filter_], facecolor = 'none', marker = 'D', label = f'{filter_}[ZP = {zp_median:.2f} ± {zp_std:.2f}]')
        plt.axhline(zp_median, color = color_map[filter_], ls = '--', lw = 1)
        # plt.axhline(zp_median + zp_std, color = 'k', ls = '--', lw = 1)
        # plt.axhline(zp_median - zp_std, color = 'k', ls = '--', lw = 1)
        # plt.text(0.05, 0.90, f'ZP = {zp_median:.3f} ± {zp_std:.3f}', transform = plt.gca().transAxes, fontsize = 14)
        # plt.ylim(zp_median - 0.5, zp_median + 0.5)
        # plt.show()    
        zp_filter[filter_] = zp_median
        zp_filter_std[filter_] = zp_std
    plt.legend(ncols = 3)
    plt.ylabel(r'ZP [M$_{PSF}$ - M$_{GAIAXP}$]', fontsize = 14)
    plt.xlabel('Magnitude [mag]', fontsize = 14)
    #plt.ylim(24, 26)
    plt.show()
    
#%%

#%% COMPARE WITH EZPHOT RESULT
if __name__ == '__main__':
    catalogset = dbrowser.search(pattern = 'coadd**.fits.cat', return_type = 'catalog')
    catalogset.select_sources(target_ra[0], target_dec[0])
    from ezphot.dataobjects import PhotometricSpectrum
    photspec = PhotometricSpectrum(catalogset)
    flux_key_ezphot = 'MAGSKY_APER_2'
    fluxerr_key_ezphot = 'MAGERR_APER_2'
    zperr_key_ezphot = 'ZPERR_APER_2'
    photspec_result = photspec.plot(ra = target_ra[0], dec = target_dec[0], flux_key = flux_key_ezphot, fluxerr_key = fluxerr_key_ezphot)
    photspec_fig = list(photspec_result[0].values())[0]
    photspec_tbl = photspec_result[3]
    photspec_ax = list(photspec_result[2].values())[0]
# %%
if __name__ == '__main__':
    result_target = result[-1]
    result_host = result[-1]
    magdiff_all = []
    for flux_key in flux_keys:
        filter_key = flux_key.replace('FLUX_', '')
        fluxerr_key = flux_key.replace('FLUX_', 'FLUXERR_')
        flux_key_fit = flux_key + '_fit'
        fluxerr_key_fit = fluxerr_key + '_fit'
        filter = flux_key.replace('FLUX_', '')
        
        flux_fit = result_target[flux_key_fit]
        fluxerr_fit = result_target[fluxerr_key_fit]
        abmag_fit = -2.5*np.log10(flux_fit) + zp_filter[filter]
        abmagerr_fit = 2.5/np.log(10) * fluxerr_fit / flux_fit
        zperr_fit = zp_filter_std[filter]
        magerr_fit = np.sqrt(abmagerr_fit**2 + zperr_fit**2)
        
        row_photspec = photspec_tbl[photspec_tbl['filter'] == filter_key]
        abmag_photspec = row_photspec[flux_key_ezphot]
        magerr_photspec = np.sqrt(row_photspec[fluxerr_key_ezphot]**2 + row_photspec[zperr_key_ezphot]**2)
        
        wl = photspec.FILTER_PIVOT_WAVELENGTH_NM[filter_key]
        mag_diff = float(abmag_fit - abmag_photspec[0])
        photspec_ax.scatter(wl, abmag_fit, edgecolor = 'red', facecolor = 'none', marker = 'D')
        photspec_ax.errorbar(wl, abmag_fit, yerr = magerr_fit, color = 'r', fmt = 'none')
        photspec_ax.text(wl, 15.0, f'{mag_diff:.3f}', color = 'r', rotation = 90)
        magdiff_all.append(mag_diff)

#%%
if __name__ == '__main__':
    print(np.mean(magdiff_all), np.std(magdiff_all), np.median(magdiff_all))
#%%
photspec_fig
#%%
# # ================================================
# # Photometry
# # ================================================
# # RUN TRACT7DT (TARGET)
# if __name__ == '__main__':
#     import numpy as np
#     self = Tract7DTRunner(image_paths = image_paths, filter_list = filter_list, catalog_paths = catalog_paths)
#     list_ra = [233.857430764]
#     list_dec = [12.0577222937]
#     list_type = ['STAR'] * len(list_ra)
#     self.register_target(list_ra = list_ra, list_dec = list_dec, list_type = list_type, update_type_from_catalog = False)
#     self.register_reference_catalog(objname = 'T22956')
#     self.run()

# #%% READ TRACT7DT RESULT
# if __name__ == '__main__':
#     from astropy.io import ascii
#     result = ascii.read(self.workdir / 'final_catalog_with_fit.csv')    
# #%% COMPARE WITH EZPHOT RESULT
# if __name__ == '__main__':
#     catalogset = dbrowser.search(pattern = 'coadd*20250520*.fits.cat', return_type = 'catalog')
#     catalogset.select_sources(233.857430764, 12.0577222937)
#     from ezphot.dataobjects import PhotometricSpectrum
#     photspec = PhotometricSpectrum(catalogset)
#     flux_key_ezphot = 'MAGSKY_APER_2'
#     fluxerr_key_ezphot = 'MAGERR_APER_2'
#     zperr_key_ezphot = 'ZPERR_APER_2'
#     photspec_result = photspec.plot(ra = 233.857430764, dec = 12.0577222937, flux_key = flux_key_ezphot, fluxerr_key = fluxerr_key_ezphot)
#     photspec_fig = photspec_result[0]['2025-05-20 02:05']
#     photspec_tbl = photspec_result[3]
#     photspec_ax = photspec_result[2]['2025-05-20 02:05']
# # %%
# if __name__ == '__main__':
#     magdiff_all = []
#     chisq_all = []
#     for flux_key in flux_keys:
#         filter_key = flux_key.replace('FLUX_', '')
#         flux_key_fit = flux_key + '_fit'
#         filter = flux_key.replace('FLUX_', '')
#         flux_fit = result[flux_key_fit]
#         abmag_fit = -2.5*np.log10(flux_fit) + zp_filter[filter]
#         row_photspec = photspec_tbl[photspec_tbl['filter'] == filter_key]
#         abmag_photspec = row_photspec[flux_key_ezphot]
#         magerr_photspec = np.sqrt(row_photspec[fluxerr_key_ezphot]**2 + row_photspec[zperr_key_ezphot]**2)
#         chisq = (abmag_fit[0] - abmag_photspec[0])**2 / magerr_photspec[0]**2
#         wl = photspec.FILTER_PIVOT_WAVELENGTH_NM[filter_key]
#         mag_diff = float(abmag_fit[0] - abmag_photspec[0])
#         photspec_ax.scatter(wl, abmag_fit[0], edgecolor = 'blue', facecolor = 'none', marker = 'D')
#         photspec_ax.text(wl, 14.3, f'{mag_diff:.3f}', color = 'blue', rotation = 90)
#         magdiff_all.append(mag_diff)
#         chisq_all.append(chisq)


#%%
from pathlib import Path
import json
from typing import Union
import shutil
import numpy as np

class Configuration:
    def __init__(self,
                 telkey: str = None,
                 configpath: Union[Path, str] = Path.home()/'ezphot/config'):#Path(__file__).resolve().parent):
        if not Path(configpath).exists():
            print(f"[CRITICAL] Configuration path {configpath} does not exist. Run initialization first.")
        self.telkey = telkey
        self.config = dict()

        # Set up paths
        self.path_home = Path.home()
        self.path_base = Path(__file__).resolve().parent.parent
        self.path_config = Path(configpath)
        self.path_ezphot =  self.path_config.parent
        self.path_log = self.path_ezphot / 'log'
        self.path_data = self.path_ezphot / 'data'
        
        # Gloabl configuration files
        self.path_config_global = self.path_config / 'common'
        self._configfiles_global = list(self.path_config_global.glob('*.config'))
        config_global = self._load_configuration(self._configfiles_global)
        self.config.update(config_global)
        
        # Telescope specific configuration files
        self.path_config_specific = self.path_config / 'specific'
        if self.telkey is not None:
            self.telinfo = self._load_telinfo()
            self.path_config_specific_telescope = self.path_config_specific / self.telkey
            self._configfiles_telescopes = list(self.path_config_specific_telescope.glob('*.config'))
            if not self._configfiles_telescopes:
                print('No configuration file is found.\n Do you want to make default configuration files? (y/n)')
                answer = input()
                if answer == 'y':
                    self.register_telescope()
                    self._configfiles_telescopes = list(self.path_config_specific_telescope.glob('*.config'))
                    config_unit = self._load_configuration(self._configfiles_telescopes)
                    self.config.update(config_unit)    
                else:
                    print("WARNING: Only common configuration files are loaded.")
            else:
                config_unit = self._load_configuration(self._configfiles_telescopes)
                self.config.update(config_unit)           
                
    def initialize(self, copy_default: bool = True):
        """Initialize the configuration by creating necessary config files."""
        # Make sure base paths exist
        self.path_config_global.mkdir(parents=True, exist_ok=True)
        self.path_config_specific.mkdir(parents=True, exist_ok=True)
        self.path_log.mkdir(parents=True, exist_ok=True)
        self.path_data.mkdir(parents=True, exist_ok=True)

        print(f"Global configuration path created: {self.path_config_global}")
        print(f"Specific configuration path created: {self.path_config_specific}")
        
        default_global_config_path = self.path_base / 'configuration' /  'common'
        default_specific_config_path = self.path_base / 'configuration' / 'specific'
        # Copy default config files to the folder
        if copy_default:
            shutil.copytree(default_global_config_path, self.path_config_global, dirs_exist_ok=True)
            print(f"Copied default global configs from {default_global_config_path}")
            shutil.copytree(default_specific_config_path, self.path_config_specific, dirs_exist_ok=True)
            print(f"Copied default telescope configs from {default_specific_config_path}")
            telescope_keys = ['\n' + p.name for p in default_specific_config_path.iterdir() if p.is_dir()]
            telescope_keys_str = ' '.join(telescope_keys)
            print(f'Current available telescope keys: {telescope_keys_str}')
            
        for tel_key in self.available_telescope_keys:
            # Rsgister each telescope
            instance = Configuration(telkey=tel_key)
            instance.path_config_specific_telescope.mkdir(parents=True, exist_ok=True)
            instance._register_telescope()
        
        self = Configuration()
        # After creating all config files, load them and make sure all directories exist
        for path in list(self.config.values()):
            self._ensure_dirs_exist(path)
            
    def register_telescope(self):
        if not self.telkey:
            raise ValueError("Telescope key (telkey) must be provided to initialize configuration.")

        self.path_config_specific_telescope.mkdir(parents=True, exist_ok=True)

        # Create config files
        self._register_telescope()
        
        for path in list(self.config.values()):
            self._ensure_dirs_exist(path)
            
    @property
    def available_telescope_keys(self):
        """Return a list of available telescope keys."""
        return [p.name for p in self.path_config_specific.iterdir() if p.is_dir()]
    
    def _load_telinfo(self):
        from astropy.io import ascii
        telinfo_tbl = ascii.read(self.path_config_global / 'observatory_info.dat', format='fixed_width')
        
        # Parse telescope key
        telkey_keys = self.telkey.split('_')
        if len(telkey_keys) == 4:
            observatory, ccd, readoutmode, binning_str = telkey_keys
        elif len(telkey_keys) == 3:
            observatory, ccd, binning_str = telkey_keys
            readoutmode = None
        else:
            raise ValueError(f"Invalid telescope key: {self.telkey}. Telescope key should be in the format of 'observatory_ccd_readoutmode_binning' or 'observatory_ccd_binning'")
        
        # Extract binning number from format like "1x1" -> 1
        binning = int(binning_str.split('x')[0])
        
        # Filter by observatory
        telinfo = telinfo_tbl[telinfo_tbl['telescope'] == observatory]
        
        # Filter by CCD
        telinfo = telinfo[telinfo['ccd'] == ccd]
        
        # Filter by binning
        telinfo = telinfo[telinfo['binning'] == binning]
        
        # Filter by readout mode if provided
        if readoutmode is not None:
            telinfo = telinfo[telinfo['readoutmode'] == readoutmode]
        
        # Return result
        if len(telinfo) == 0:
            raise ValueError(f"No telescope information found for {self.telkey} in observatory_info.dat. \n Please follow the below instructions. \n 1. Enter necessary telescope information to {self.path_config_global / 'observatory_info.dat'} \n 2. Register by running Configuration(telkey).register_telescope()")
        else:
            return telinfo[0]
        
    def _load_configuration(self, configfiles):
        all_config = dict()
        for configfile in configfiles:
            with open(configfile, 'r') as f:
                config = json.load(f)
                all_config.update(config)
        return all_config

    def _make_configfile(self, dict_params: dict, filename: str, savepath: Union[str, Path]):
        filepath = Path(savepath) / filename
        with open(filepath, 'w') as f:
            json.dump(dict_params, f, indent=4)
        print(f'New configuration file made: {filepath}')

    def _ensure_dirs_exist(self, *paths):
        """Create all directories in the given paths if they don't exist."""
        for p in paths:
            try:
                Path(p).mkdir(parents=True, exist_ok=True)
            except Exception as e:
                pass
                #print(f"[WARNING] Could not create directory {p}: {e}")
                
    def _register_telescope(self):
        
        # Specific telescope configuration
        sex_config = dict(
            SEX_CONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.sexconfig'),
            SEX_DIR = str(self.path_config_global / 'sextractor'),
            SEX_LOGDIR = str(self.path_log / 'sextractor' / 'log'),
            SEX_HISTORYDIR = str(self.path_log / 'sextractor' / 'history')
        )
        astrometry_config = dict(
            ASTROMETRY_SEXCONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.astrometry.sexconfig')
        )
        scamp_config = dict(
            SCAMP_CONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.scampconfig'),
            SCAMP_SEXCONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.scamp.sexconfig'),
            SCAMP_DIR = str(self.path_config_global / 'scamp'),
            SCAMP_LOGDIR = str(self.path_log / 'scamp' / 'log'),
            SCAMP_HISTORYDIR = str(self.path_log / 'scamp' / 'history')
        )
        swarp_config = dict(
            SWARP_CONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.swarpconfig'),
            SWARP_DIR = str(self.path_config_global / 'swarp'),
            SWARP_LOGDIR = str(self.path_log / 'swarp' / 'log'),
            SWARP_HISTORYDIR = str(self.path_log / 'swarp' / 'history')
        )
        psfex_config = dict(
            PSFEX_CONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.psfexconfig'),
            PSFEX_SEXCONFIG = str(self.path_config_specific_telescope / f'{self.telkey}.psfex.sexconfig'),
            PSFEX_DIR = str(self.path_config_global / 'psfex'),
            PSFEX_LOGDIR = str(self.path_log / 'psfex' / 'log'),
            PSFEX_HISTORYDIR = str(self.path_log / 'psfex' / 'history')
        )
        
        for cfg, name in [
            (sex_config, 'sex.config'),
            (astrometry_config, 'astrometry.config'),
            (scamp_config, 'scamp.config'),
            (swarp_config, 'swarp.config'),
            (psfex_config, 'psfex.config')
        ]:
            self._make_configfile(cfg, name, self.path_config_specific_telescope)
        
        # Update default configuration files to the telescope specific ones
        default_sexconfig = self.path_config_global / 'default.sexconfig'
        default_astrometrysexconfig = self.path_config_global / 'default.astrometry.sexconfig'
        default_swarpconfig = self.path_config_global / 'default.swarpconfig'
        default_scampconfig = self.path_config_global / 'default.scampconfig'
        default_scampsexconfig = self.path_config_global / 'default.scamp.sexconfig'
        default_psfexconfig = self.path_config_global / 'default.psfexconfig'
        default_psfexsexconfig = self.path_config_global / 'default.psfex.sexconfig'
        self._update_default_sexconfig(default_sexconfig, sex_config['SEX_CONFIG'])
        self._update_default_sexconfig(default_astrometrysexconfig, astrometry_config['ASTROMETRY_SEXCONFIG'])
        self._update_default_swarpconfig(default_swarpconfig, swarp_config['SWARP_CONFIG'])
        self._update_default_scampconfig(default_scampconfig, scamp_config['SCAMP_CONFIG'])
        self._update_default_sexconfig(default_scampsexconfig, scamp_config['SCAMP_SEXCONFIG'])
        self._update_default_psfexconfig(default_psfexconfig, psfex_config['PSFEX_CONFIG'])
        self._update_default_sexconfig(default_psfexsexconfig, psfex_config['PSFEX_SEXCONFIG'])

        # Global configuration
        calibdata_config = dict(
            CALIBDATA_DIR = str(self.path_data / 'calibdata'),
            CALIBDATA_MASTERDIR = str(self.path_data / 'mcalibdata'),
            )
        refdata_config = dict(
            REFDATA_DIR = str(self.path_data / 'refdata'),
            )
        obsdata_config = dict(OBSDATA_DIR = str(self.path_data / 'obsdata'))
        scidata_config = dict(SCIDATA_DIR = str(self.path_data / 'scidata'))
        catalog_config = dict(CATALOG_DIR = str(self.path_data / 'skycatalog' / 'archive'))
        observatory_config = dict(
            OBSERVATORY_LOCATIONINFO = str(self.path_config_global / 'obs_location.txt'),
            OBSERVATORY_TELESCOPEINFO = str(self.path_config_global / 'observatory_info.dat'),
            OBSERVATORY_TELESCOPEHINT = str(self.path_config_global / 'observatory_info_hint.yaml')
        )
        
        sdtdata_config = dict(
            SDTDATA_OBSSOURCEDIR = str(self.path_data / 'connecteddata' / '7DT' / 'obsdata'),
            SDTDATA_OBSDESTDIR = str(self.path_data / 'obsdata' / '7DT'),
            SDTDATA_SCISOURCEDIR = str(self.path_data / 'connecteddata' / '7DT' / 'processed_1x1_gain2750'),
            SDTDATA_SCIDESTDIR = str(self.path_data / 'scidata' / '7DT' / '7DT_C361K_HIGH_1x1')
        )
        
        synphot_config = dict(SYNPHOT_FILTERDIR = str(self.path_config_global / 'transmission'))

        for cfg, name in [
            (calibdata_config, 'calibdata.config'),
            (refdata_config, 'refdata.config'),
            (scidata_config, 'scidata.config'),
            (catalog_config, 'catalog.config'),
            (observatory_config, 'observatory.config'),
            (sdtdata_config, 'sdtdata.config'),
            (obsdata_config, 'obsdata.config'),
            (synphot_config, 'synphot.config')
        ]:
            self._make_configfile(cfg, name, self.path_config_global)

        # Remove per-telescope specific keys before saving global versions
        sex_config['SEX_CONFIG'] = str(default_sexconfig)
        scamp_config['SCAMP_CONFIG'] = str(default_scampconfig)
        scamp_config['SCAMP_SEXCONFIG'] = str(default_scampsexconfig)
        swarp_config['SWARP_CONFIG'] = str(default_swarpconfig)
        psfex_config['PSFEX_CONFIG'] = str(default_psfexconfig)
        psfex_config['PSFEX_SEXCONFIG'] = str(default_psfexsexconfig)
        astrometry_config['ASTROMETRY_SEXCONFIG'] = str(default_astrometrysexconfig)

        for cfg, name in [
            (sex_config, 'sex.config'),
            (scamp_config, 'scamp.config'),
            (swarp_config, 'swarp.config'),
            (psfex_config, 'psfex.config'),
            (astrometry_config, 'astrometry.config'),
        ]:
            self._make_configfile(cfg, name, self.path_config_global)
    
    def _update_default_sexconfig(self, default_path: Union[str, Path] = None, output_path: Union[str, Path] = None):
        
        telinfo = self.telinfo
        if default_path is None:
            default_path = self.path_config_global / 'default.sexconfig'
        if output_path is None:
            output_path = self.path_config_specific_telescope / f'{self.telkey}.sexconfig'
        # Define updates based on telescope info
        update_dict = {}
        if telinfo is not None:
            # Update based on telescope properties
            update_dict['GAIN'] = telinfo['gain']
            update_dict['READNOISE'] = telinfo['readnoise']
            update_dict['PIXEL_SCALE'] = telinfo['pixelscale']
            update_dict['DETECT_MINAREA'] = int(np.pi* (2/telinfo['pixelscale']/2)**2) if telinfo['pixelscale'] < 1 else int(np.pi* (4/telinfo['pixelscale']/2)**2)
            update_dict['SEEING_FWHM'] = 2.5 if telinfo['pixelscale'] < 1 else 4
            update_dict['PHOT_APERTURES'] = f'{round(5 / telinfo["pixelscale"],2)},{round(7 / telinfo["pixelscale"],2)},{round(10 / telinfo["pixelscale"],2)}'
            if update_dict['DETECT_MINAREA'] < 3:
                update_dict['DETECT_MINAREA'] = 3

        self._update_defulatconfig(update_dict, default_path, output_path)

    def _update_default_swarpconfig(self, default_path: Union[str, Path] = None, output_path: Union[str, Path] = None):
        
        telinfo = self.telinfo
        if default_path is None:
            default_path = self.path_config_global / 'default.swarpconfig'
        if output_path is None:
            output_path = self.path_config_specific_telescope / f'{self.telkey}.swarpconfig'
        # Define updates based on telescope info
        update_dict = {}
        if telinfo is not None:
            # Update based on telescope properties
            update_dict['GAIN_DEFAULT'] = telinfo['gain']
            update_dict['PIXEL_SCALE'] = telinfo['pixelscale']
            update_dict['IMAGE_SIZE'] = f'{round(1.1*telinfo["x"],-2)},{round(1.1*telinfo["y"],-2)}'

        self._update_defulatconfig(update_dict, default_path, output_path)

    def _update_default_scampconfig(self, default_path: Union[str, Path] = None, output_path: Union[str, Path] = None):
        
        if default_path is None:
            default_path = self.path_config_global / 'default.scampconfig'
        if output_path is None:
            output_path = self.path_config_specific_telescope / f'{self.telkey}.scampconfig'
        
        update_dict = dict()
        self._update_defulatconfig(update_dict, default_path, output_path)
        
    def _update_default_psfexconfig(self, default_path: Union[str, Path] = None, output_path: Union[str, Path] = None):
        
        if default_path is None:
            default_path = self.path_config_global / 'default.psfexconfig'
        if output_path is None:
            output_path = self.path_config_specific_telescope / f'{self.telkey}.psfexconfig'
            
        update_dict = dict()
        self._update_defulatconfig(update_dict, default_path, output_path)
    
    def _update_defulatconfig(self, 
                              update_dict: dict, 
                              default_path: Union[str, Path], 
                              output_path: Union[str, Path]):

        """Update specific keywords in the default.sexconfig based on telescope info."""
        
        # Read the original file
        config_path = default_path
        with open(config_path, 'r') as f:
            lines = f.readlines()
        
        # Update the lines
        updated_lines = []
        for line in lines:
            # Skip comments and empty lines
            if line.strip().startswith('#') or not line.strip():
                updated_lines.append(line)
                continue
            
            # Check if this line contains a key we want to update
            parts = line.split()
            if len(parts) >= 2:
                key = parts[0]
                if key in update_dict:
                    # Replace the value while preserving formatting
                    value = update_dict[key]
                    # Keep the original spacing and comments
                    comment_start = line.find('#')
                    if comment_start != -1:
                        comment = line[comment_start:]
                        # Format: KEY    VALUE    # comment
                        updated_line = f"{key:<15} {value:<10} {comment}\n"
                    else:
                        updated_line = f"{key:<15} {value}\n"
                    updated_lines.append(updated_line)
                else:
                    updated_lines.append(line)
            else:
                updated_lines.append(line)
                        
        # Ensure directory exists
        Path(output_path).parent.mkdir(parents=True, exist_ok=True)
        
        # Write the updated configuration
        with open(output_path, 'w') as f:
            f.writelines(updated_lines)
        
        print("New configuration file made: ", output_path)

#%%
if __name__ == '__main__':
    # For the first time, initiailize the configuration
    self = Configuration()
    #self.initialize(copy_default=True)
    
    telescope_keys = [
        '7DT_C361K_HIGH_1x1', '7DT_C361K_HIGH_2x2', '7DT_C361K_LOW_1x1', '7DT_C361K_LOW_2x2',
        'CBNUO_STX16803_1x1', 'LSGT_SNUCAMII_1x1', 'LSGT_ASI1600MM_1x1',
        'RASA36_KL4040_HIGH_1x1', 'RASA36_KL4040_MERGE_1x1', 'SAO_C361K_1x1',
        'SOAO_FLI4K_1x1', 'KCT_STX16803_1x1']

    #self = Configuration(telkey='7DT_C361K_HIGH_1x1')
    for key in telescope_keys:
        print(key)
        
        config = Configuration(telkey=key)
        config.register_telescope()
        print(config.config)
#%

# %%



import numpy as np
import os
from astropy.coordinates import SkyCoord
import astropy.units as u
from astropy.wcs import WCS

import matplotlib.pyplot as plt
plt.ioff()
import matplotlib as mpl
mpl.rcParams["axes.titlesize"] = 14
mpl.rcParams["axes.labelsize"] = 20
plt.rcParams['savefig.dpi'] = 500
plt.rc('font', family='serif')

# When downloading images, coordinates in degrees is needed, and when using Swarp, coordinates in HMS DMS is needed
# If both format is prepared, no need of this code
# Just converting coordinate format
def degrees_to_hms_dms(ra_in_degrees, dec_in_degrees):
    """
    Convert RA in degrees to HMS (hours, minutes, seconds) and 
    Dec in degrees to DMS (degrees, minutes, seconds).
    
    Parameters:
    ra_in_degrees (float): RA in degrees.
    dec_in_degrees (float): Dec in degrees.

    Returns:
    tuple: The RA in HMS format (HH:MM:SS) and Dec in DMS format (±DD:MM:SS.ssss).
    """
    coord = SkyCoord(ra=ra_in_degrees * u.deg, dec=dec_in_degrees * u.deg, frame='icrs')
    
    ra_hms = coord.ra.to_string(unit=u.hourangle, sep=':', pad=True, precision=0)  # HH:MM:SS
    dec_dms = coord.dec.to_string(unit=u.degree, sep=':', alwayssign=True, precision=4)  # ±DD:MM:SS.ssss
    
    return ra_hms, dec_dms

def generate_pointings(center_ra, center_dec, xsize, ysize, pixscale, n=8, m=8, margin_frac=0.1):
    """
    Generate RA, Dec pointings based on input center coordinates and image size.
    
    Parameters:
    - center_ra (float): Central RA coordinate in degrees.
    - center_dec (float): Central Dec coordinate in degrees.
    - xsize (int): X size of the image in pixels.
    - ysize (int): Y size of the image in pixels.
    - pixscale (float): Pixel scale in arcseconds/pixel.
    - n (int): Number of pointings along the X-axis (default is 8).
    - m (int): Number of pointings along the Y-axis (default is 8).
    - margin_frac (float): Fractional margin to add around the image (default is 10%).
    
    Returns:
    - List of tuples with (RA, Dec) for each pointing.
    """
    
    # Calculate total margin to cover the entire field of view
    xmargin = xsize * (1 + margin_frac)
    ymargin = ysize * (1 + margin_frac)

    # Convert pixel margins to degrees
    ra_range = xmargin * pixscale / 3600.0 / 2.0  # Convert arcsec to degrees
    dec_range = ymargin * pixscale / 3600.0 / 2.0

    # Generate evenly spaced RA and Dec points
    ra_points = np.linspace(center_ra - ra_range, center_ra + ra_range, n)
    dec_points = np.linspace(center_dec - dec_range, center_dec + dec_range, m)

    # Create all combinations of RA and Dec pointings
    pointings = [(ra, dec) for ra in ra_points for dec in dec_points]

    return pointings

def plot_rectangle_on_sky(center_ra, center_dec, x_size, y_size, pixel_scale, color='red'):
    """
    Plot a rectangle on the sky with a given center and size in arcseconds.
    
    Parameters:
    - center_ra (float): Central RA in degrees.
    - center_dec (float): Central Dec in degrees.
    - x_size (int): X size of the image in pixels.
    - y_size (int): Y size of the image in pixels.
    - pixel_scale (float): Pixel scale in arcseconds per pixel.
    - color (str): Color of the rectangle.
    """
    # Convert size from pixels to degrees
    x_deg = x_size * pixel_scale / 3600.0
    y_deg = y_size * pixel_scale / 3600.0
    
    # Plot the rectangle with four corners
    ra_min = center_ra - x_deg / 2
    ra_max = center_ra + x_deg / 2
    dec_min = center_dec - y_deg / 2
    dec_max = center_dec + y_deg / 2
    
    # Plot the rectangle (as a series of lines between the corners)
    plt.plot([ra_min, ra_max], [dec_min, dec_min], color=color)  # Bottom edge
    plt.plot([ra_min, ra_max], [dec_max, dec_max], color=color)  # Top edge
    plt.plot([ra_min, ra_min], [dec_min, dec_max], color=color)  # Left edge
    plt.plot([ra_max, ra_max], [dec_min, dec_max], color=color)  # Right edge

def plot_pointings_on_sky(pointings, center_ra, center_dec, x_size, y_size, pixel_scale, target_outpath, title=None, save=True):
    """
    Plot the pointings and the central rectangle on the sky.
    
    Parameters:
    - pointings (list of tuples): List of (RA, Dec) for each pointing.
    - center_ra (float): Central RA in degrees.
    - center_dec (float): Central Dec in degrees.
    - x_size (int): X size of the image in pixels.
    - y_size (int): Y size of the image in pixels.
    - pixel_scale (float): Pixel scale in arcseconds per pixel.
    - target_outpath (str): Path to save the plot image.
    - title (str, optional): Title of the plot.
    - save (bool, optional): Whether to save the plot (default is True).
    """
    
    # Calculate the aspect ratio from x_size and y_size
    aspect_ratio = x_size / y_size
    
    # Multiply by a scaling factor (4 in this case) to make the figure large enough
    scaling_factor = 4
    figsize_x = aspect_ratio * scaling_factor
    figsize_y = scaling_factor
    
    # Set the figure size based on the aspect ratio
    plt.figure(figsize=(figsize_x, figsize_y))

    # Plot all pointings
    ra_vals, dec_vals = zip(*pointings)
    plt.plot(ra_vals, dec_vals, marker='.', ls='none', color='grey', label='Pointings')
    
    # Plot the center and the rectangle
    plt.plot(center_ra, center_dec, 'rx', ms=10, label='Center')
    plot_rectangle_on_sky(center_ra, center_dec, x_size, y_size, pixel_scale, color='red')
    
    # Labels and title
    plt.xlabel("RA (deg)")
    plt.ylabel("Dec (deg)")
    if title:
        plt.title(title)

    # Legend
    plt.legend(loc='upper center')
    plt.tight_layout()
    # Save or show the plot
    if save:
        plt.savefig(target_outpath)
        print(f"Plot saved to {target_outpath}")
    # else:
        # plt.show()
    plt.show()
    plt.close('all')


#%%

import pandas as pd
from pathlib import Path
from typing import Dict, List, Optional
from astropy.coordinates import SkyCoord
import astropy.units as u
import numpy as np
from astropy.table import Table
from tqdm import tqdm
#%%

class Formatter:

    def __init__(self,
                 image_paths: List[Path],
                 filter_list: List[str],
                 catalog_paths: Optional[List[Path]] = None):

        self.image_paths = [Path(p) for p in image_paths]
        self.filter_list = filter_list

        if len(self.filter_list) != len(self.image_paths):
            raise ValueError("filter_list and image_paths must have same length")

        self.catalog_paths = None
        if catalog_paths is not None:
            if len(catalog_paths) != len(filter_list):
                raise ValueError("catalog_paths and filter_list must have same length")
            self.catalog_paths = [Path(p) for p in catalog_paths]

        self.prepare()

    # --------------------------------------------------------
    # Prepare
    # --------------------------------------------------------

    def prepare(self):

        # images
        image_paths = []
        filter_list = []
        for filt, img_path in zip(self.filter_list, self.image_paths):

            if not img_path.exists():
                raise FileNotFoundError(f"Image not found: {img_path}")

            # if filt in self.image_dict:
            #     raise ValueError(f"Duplicate filter: {filt}")
            image_paths.append(str(img_path))
            filter_list.append(filt)

        # catalogs (optional)
        catalog_paths = []
        if self.catalog_paths is not None:

            for filt, cat_path in zip(self.filter_list, self.catalog_paths):

                if not cat_path.exists():
                    raise FileNotFoundError(f"Catalog not found: {cat_path}")

                catalog_paths.append(str(cat_path))
        
        self.image_paths = image_paths
        self.filter_list = filter_list
        self.catalog_paths = catalog_paths

    # --------------------------------------------------------
    # 1️⃣ Return images
    # --------------------------------------------------------

    def to_input_images(self, output_path: Path):
        """
        Save image paths to text file in filter_list order.

        Output format:
            one image path per line
        """

        if not self.image_paths:
            raise RuntimeError("image_dict is empty. Populate it before saving.")

        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        with open(output_path, "w") as f:
            for image_path in self.image_paths:
                f.write(str(image_path) + "\n")
        
        return str((output_path).resolve())
        
    # --------------------------------------------------------
    # 2️⃣ Match catalogs to targets
    # --------------------------------------------------------
    def to_input_catalogs(self,
                          output_path: Path,
                          list_ra: List[float],
                          list_dec: List[float],
                          list_type: List[str] = None,
                          list_ellip: List[float] = None,
                          list_Re: List[float] = None,
                          list_theta: List[float] = None,
                          
                          update_type_from_catalog: bool = True,
                          update_flux_from_catalog: bool = True,
                          update_ellip_from_catalog: bool = True,
                          update_Re_from_catalog: bool = True,
                          update_theta_from_catalog: bool = True,
                          
                          ra_key: str = 'X_WORLD',
                          dec_key: str = 'Y_WORLD',
                          type_key: str = 'CLASS_STAR',
                          flux_key: str = 'FLUX_AUTO', # Optional
                          ellip_key: str = 'ELLIPTICITY', #Optional
                          Re_key: str = 'FLUX_RADIUS', #Optional
                          theta_key: str = 'THETA_IMAGE', #Optional
                          matching_radius_arcsec: float = 5):

        """
        Match input targets to catalog entries.

        Parameters
        ----------
        list_ra : list of target RA (deg)
        list_dec : list of target DEC (deg)
        radius_arcsec : matching radius

        Returns
        -------
        dict:
            {
                filter_name: matched_flux_array
            }

        If no catalog provided:
            return None
        """
        def read_catalog(cat_path):
            """Lazy-load table data from path by trying multiple formats."""
            try_formats = [
                'fits',
                'ascii.sextractor',
                'ascii',
                'csv',
                'ascii.basic',
                'ascii.commented_header',
                'ascii.tab',
                'ascii.fast_no_header',
            ]
            for fmt in try_formats:
                try:
                    data = Table.read(cat_path, format=fmt)
                    return data  # Success
                except Exception:
                    continue  # Try next format
            return None  
        
        if len(list_ra) != len(list_dec):
            raise ValueError("list_ra and list_dec must have same length")
        if list_type is not None and len(list_type) != len(list_ra):
            raise ValueError("list_type and list_ra must have same length")
        
        n = len(list_ra)
        list_ra = np.atleast_1d(list_ra)
        list_dec = np.atleast_1d(list_dec)
        
        # Convert optional lists to numpy
        list_ellip = np.asarray(list_ellip) if list_ellip is not None else None
        list_Re = np.asarray(list_Re) if list_Re is not None else None
        list_theta = np.asarray(list_theta) if list_theta is not None else None
        list_type = np.asarray(list_type) if list_type is not None else None
        # Check lengths of lists
        if list_ellip is not None and len(list_ellip) != n:
            raise ValueError("list_ellip must have same length as list_ra and list_dec")
        if list_Re is not None and len(list_Re) != n:
            raise ValueError("list_Re must have same length as list_ra and list_dec")
        if list_theta is not None and len(list_theta) != n:
            raise ValueError("list_theta must have same length as list_ra and list_dec")
        if list_type is not None and len(list_type) != n:
            raise ValueError("list_type must have same length as list_ra and list_dec")

        target_coord = SkyCoord(list_ra, list_dec, unit="deg")
        radius = matching_radius_arcsec * u.arcsec

        result = {}
        for filt, cat_path in tqdm(zip(self.filter_list, self.catalog_paths), desc = 'Constructing input catalogs...'):

            catalog_tbl = read_catalog(cat_path)
            if catalog_tbl is None:
                print(f"[CRITICAL] {filt}: Failed to read catalog")
                continue

            if ra_key not in catalog_tbl.colnames or dec_key not in catalog_tbl.colnames:
                print(f"[CRITICAL] {filt}: Missing RA/DEC")
                continue

            cat_coord = SkyCoord(catalog_tbl[ra_key],
                                catalog_tbl[dec_key],
                                unit="deg")

            idx, sep2d, _ = target_coord.match_to_catalog_sky(cat_coord)
            matched = sep2d < radius

            # Initialize arrays depending on update flags
            flux_array = np.full(n, np.nan)
            ellip_array = np.full(n, np.nan)
            Re_array = np.full(n, np.nan)
            theta_array = np.full(n, np.nan)
            type_array = np.full(n, 'EXP', dtype=object)

            # ---- If update is False → use provided lists ----
            if not update_ellip_from_catalog and list_ellip is not None:
                ellip_array = list_ellip.copy()

            if not update_Re_from_catalog and list_Re is not None:
                Re_array = list_Re.copy()

            if not update_theta_from_catalog and list_theta is not None:
                theta_array = list_theta.copy()

            if not update_type_from_catalog and list_type is not None:
                type_array = list_type.copy()

            # ---- Update from catalog if requested ----
            for i, (is_match, cat_index) in enumerate(zip(matched, idx)):

                if not is_match:
                    continue

                row = catalog_tbl[cat_index]

                if update_flux_from_catalog and flux_key in catalog_tbl.colnames:
                    flux_array[i] = row[flux_key]

                if update_ellip_from_catalog and ellip_key in catalog_tbl.colnames:
                    ellip_array[i] = row[ellip_key]

                if update_Re_from_catalog and Re_key in catalog_tbl.colnames:
                    Re_array[i] = row[Re_key]

                if update_theta_from_catalog and theta_key in catalog_tbl.colnames:
                    theta_array[i] = row[theta_key]

                if update_type_from_catalog and type_key in catalog_tbl.colnames:
                    type_value = row[type_key]
                    type_array[i] = 'STAR' if type_value > 0.5 else 'EXP'

            result[filt] = {
                "flux": flux_array,
                "ellip": ellip_array,
                "Re": Re_array,
                "theta": theta_array,
                "type": type_array
            }
            self.build_input_catalog(output_path = output_path,
                                      list_ra = list_ra,
                                      list_dec = list_dec,
                                      result_dict = result,
                                      list_id = None)

        return result

    def build_input_catalog(self,
                            output_path: Path,
                            list_ra: List[float],
                            list_dec: List[float],
                            result_dict: Dict,
                            list_id: Optional[List[str]] = None):

        n = len(list_ra)

        if list_id is None:
            list_id = [f"{i:05d}" for i in range(n)]

        df = pd.DataFrame({
            "ID": list_id,
            "RA": list_ra,
            "DEC": list_dec
        })

        # ==========================
        # TYPE → most probable (mode)
        # ==========================

        type_array = []

        for i in range(n):
            types_i = []

            for filt in self.filter_list:
                if filt in result_dict:
                    t = result_dict[filt]["type"][i]
                    if t is not None:
                        types_i.append(t)

            if len(types_i) == 0:
                type_array.append("EXP")
            else:
                # mode
                vals, counts = np.unique(types_i, return_counts=True)
                type_array.append(vals[np.argmax(counts)])

        df["TYPE"] = type_array

        # ==========================
        # Flux columns (unchanged)
        # ==========================

        for filt in self.filter_list:
            colname = f"FLUX_{filt}"
            if filt in result_dict:
                df[colname] = result_dict[filt]["flux"]
            else:
                df[colname] = np.nan

        # ==========================
        # Shape parameters
        # ==========================

        ell_array = []
        Re_array = []
        theta_array = []

        for i in range(n):

            ell_vals = []
            Re_vals = []
            theta_vals = []

            for filt in self.filter_list:
                if filt not in result_dict:
                    continue

                ell = result_dict[filt]["ellip"][i]
                Re_val = result_dict[filt]["Re"][i]
                theta = result_dict[filt]["theta"][i]

                if not np.isnan(ell):
                    ell_vals.append(ell)

                if not np.isnan(Re_val):
                    Re_vals.append(Re_val)

                if not np.isnan(theta):
                    theta_vals.append(theta)

            # ---- ELL median ----
            ell_array.append(np.nanmedian(ell_vals) if len(ell_vals) > 0 else np.nan)

            # ---- Re median ----
            Re_array.append(np.nanmedian(Re_vals) if len(Re_vals) > 0 else np.nan)

            # ---- THETA circular mean ----
            if len(theta_vals) > 0:
                theta_rad = np.deg2rad(theta_vals)
                mean_angle = np.arctan2(
                    np.nanmean(np.sin(theta_rad)),
                    np.nanmean(np.cos(theta_rad))
                )
                theta_array.append(np.rad2deg(mean_angle))
            else:
                theta_array.append(np.nan)

        df["ELL"] = ell_array
        df["Re"] = Re_array
        df["THETA"] = theta_array

        # ---- Save ----
        output_path = Path(output_path)
        output_path.parent.mkdir(parents=True, exist_ok=True)

        df.to_csv(output_path, index=False)

        return str(output_path.resolve())



    # def build_input_catalog(self,
    #                         output_path: Path,
    #                         list_ra: List[float],
    #                         list_dec: List[float],
    #                         result_dict: Dict,
    #                         list_id: Optional[List[str]] = None):
    #     """
    #     Construct final tract7DT input catalog file.
    #     """

    #     n = len(list_ra)

    #     if list_id is None:
    #         list_id = [f"{i:05d}" for i in range(n)]

    #     # ---- Build DataFrame ----
    #     df = pd.DataFrame({
    #         "ID": list_id,
    #         "RA": list_ra,
    #         "DEC": list_dec
    #     })

    #     # ---- TYPE (global) ----
    #     type_array = np.full(n, "EXP", dtype=object)

    #     for filt in result_dict:
    #         band_type = result_dict[filt]["type"]
    #         for i in range(n):
    #             if band_type[i] is not None and band_type[i] != "EXP":
    #                 type_array[i] = band_type[i]

    #     df["TYPE"] = type_array

    #     # ---- Flux columns ----
    #     # Use filter_list order
    #     for filt in self.filter_list:
    #         colname = f"FLUX_{filt}"

    #         if filt in result_dict:
    #             df[colname] = result_dict[filt]["flux"]
    #         else:
    #             df[colname] = np.nan

    #     # ---- Shape parameters (take first valid band) ----
    #     ell_array = np.full(n, np.nan)
    #     Re_array = np.full(n, np.nan)
    #     theta_array = np.full(n, np.nan)

    #     for filt in self.filter_list:
    #         if filt not in result_dict:
    #             continue

    #         ell_band = result_dict[filt]["ellip"]
    #         Re_band = result_dict[filt]["Re"]
    #         theta_band = result_dict[filt]["theta"]

    #         for i in range(n):
    #             if np.isnan(ell_array[i]) and not np.isnan(ell_band[i]):
    #                 ell_array[i] = ell_band[i]
    #                 Re_array[i] = Re_band[i]
    #                 theta_array[i] = theta_band[i]

    #     df["ELL"] = ell_array
    #     df["Re"] = Re_array
    #     df["THETA"] = theta_array

    #     # ---- Save ----
    #     output_path = Path(output_path)
    #     output_path.parent.mkdir(parents=True, exist_ok=True)

    #     df.to_csv(output_path, index=False)

    #     return str(output_path.resolve())


    # --------------------------------------------------------
    # Summary
    # --------------------------------------------------------

    def summary(self):
        print("Tract7DT Formatter")
        print("-------------------")
        print("Filters:", self.filter_list)
        print("N bands:", len(self.filter_list))
        print("Catalog loaded:", self.catalog_paths is not None)

# %%
if __name__ == '__main__':
    from ezphot.utils import DataBrowser
    dbrowser = DataBrowser('scidata')
    dbrowser.objname = 'T22956'
    target_imgset = dbrowser.search(pattern = 'coadd*20250720*.fits', return_type = 'science')
    image_paths = [img.path for img in target_imgset.target_images]
    catalog_paths = [img.catalog.path for img in target_imgset.target_images]
    filter_list = [img.filter for img in target_imgset.target_images]
#%%
if __name__ == '__main__':
    self = Formatter(image_paths = image_paths, filter_list = filter_list, catalog_paths = catalog_paths)
# %%
# # %%

# list_ra = [233.857430764]
# list_dec = [+12.0577222937]

# # %%
# list_flux = None
# list_ellip = None
# list_Re = None
# list_theta = None
# list_type = None

# update_flux = True
# update_ellip = True
# update_Re = True
# update_theta = True
# update_type = True

# ra_key = 'X_WORLD'
# dec_key = 'Y_WORLD'
# type_key = 'CLASS_STAR'
# flux_key = 'FLUX_APER_2'
# ellip_key = 'ELLIPTICITY'
# Re_key = 'FLUX_RADIUS'
# theta_key = 'THETA_IMAGE'
# #%%
# self.to_input_images(output_path = Path('input_images.txt'))
# # %%
# self.to_input_catalogs(output_path = Path('input_catalogs.csv'),
#                        list_ra = list_ra,
#                        list_dec = list_dec,
#                        list_type = list_type,
#                        list_flux = list_flux,
#                        list_ellip = list_ellip,
#                        list_Re = list_Re,
#                        list_theta = list_theta,
#                        update_type = update_type,
#                        update_flux = update_flux,
#                        update_ellip = update_ellip,
#                        update_Re = update_Re,
#                        update_theta = update_theta,
#                        ra_key = ra_key,
#                        dec_key = dec_key,
#                        type_key = type_key,
#                        flux_key = flux_key,
#                        ellip_key = ellip_key,
#                        Re_key = Re_key,
#                        theta_key = theta_key,
#                        matching_radius_arcsec = 5)


# # %%




#%%
from ezphot.imageobjects import Mask, Errormap, Background, CalibrationImage, MasterImage
from typing import Union, List
from astropy.time import Time
from typing import Optional
import numpy as np
#%%

class ImageMethod:    
    def correct_bias(self,
                     bias_image: Union[CalibrationImage, MasterImage] = None,
                     save: bool = False,
                     verbose: bool = True
                     ):
        
        """
        Correct bias in the image.
        
        Parameters
        ----------
        bias_image : CalibrationImage or MasterImage
            The bias image to correct bias.
        save : bool, optional
            Whether to save the corrected image.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        from ezphot.methods import Preprocess
        if bias_image is None:
            bias_image = self.get_masterframe(imagetyp = 'BIAS', max_days = 100)
        preprocess = Preprocess()
        calib_image = preprocess.correct_bias(
            target_img = self,
            bias_image = bias_image,
            save = save,
            verbose = verbose
        )
        return calib_image
    
    def correct_dark(self,
                     dark_image: Union[CalibrationImage, MasterImage] = None, 
                     save : bool = False,
                     verbose: bool = True
                     ):
        """
        Correct dark in the image.
        
        Parameters
        ----------
        dark_image : CalibrationImage or MasterImage
            The dark image to correct dark.
        save : bool, optional
            Whether to save the corrected image.
        
        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        from ezphot.methods import Preprocess
        if dark_image is None:
            dark_image = self.get_masterframe(imagetyp = 'DARK', max_days = 100)
        preprocess = Preprocess()
        calib_image = preprocess.correct_dark(
            target_img = self,
            dark_image = dark_image,
            save = save,
            verbose = verbose
        )
        return calib_image
    
    def correct_flat(self,
                     flat_image: Union[CalibrationImage, MasterImage] = None,
                     save: bool = False,
                     verbose: bool = True
                     ):
        """
        Correct flat in the image.
        
        Parameters
        ----------
        flat_image : CalibrationImage or MasterImage    
            The flat image to correct flat.
        save : bool, optional
            Whether to save the corrected image.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        from ezphot.methods import Preprocess
        if flat_image is None:
            flat_image = self.get_masterframe(imagetyp = 'FLAT', max_days = 100)
        preprocess = Preprocess()
        calib_image = preprocess.correct_flat(
            target_img = self,
            flat_image = flat_image,
            save = save,
            verbose = verbose,
        )
        return calib_image
    
    def correct_bdf(self,
                    bias_image: Union[CalibrationImage, MasterImage] = None,
                    dark_image: Union[CalibrationImage, MasterImage] = None,
                    flat_image: Union[CalibrationImage, MasterImage] = None,
                    save: bool = False,
                    verbose: bool = True
                    ):
        """
        Correct bias, dark, and flat.
        
        Parameters
        ----------
        bias_image : CalibrationImage or MasterImage
            The bias image to correct bias, dark, and flat.
        dark_image : CalibrationImage or MasterImage
            The dark image to correct bias, dark, and flat.
        flat_image : CalibrationImage or MasterImage
            The flat image to correct bias, dark, and flat.
        save : bool, optional
            Whether to save the corrected image.

        Returns
        -------
        calib_image : ScienceImage
            The corrected image.
        """
        from ezphot.methods import Preprocess
        if bias_image is None:
            bias_image = self.get_masterframe(imagetyp = 'BIAS', max_days = 100)
        if dark_image is None:
            dark_image = self.get_masterframe(imagetyp = 'DARK', max_days = 100)
        if flat_image is None:
            flat_image = self.get_masterframe(imagetyp = 'FLAT', max_days = 100)
        preprocess = Preprocess()
        calib_image = preprocess.correct_bdf(
            target_img = self,
            bias_image = bias_image,
            dark_image = dark_image,
            flat_image = flat_image,
            save = save,
            verbose = verbose
        )
        return calib_image
    
    def platesolve(self,
                   overwrite: bool = True,
                   verbose: bool = True,
                   scamp_sexparams: dict = None,
                   scamp_params: dict = None
                   ):
        """
        Solve astrometry using Astrometry.net.
        
        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage]
            The target image to solve astrometry for.
        overwrite : bool, optional
            Whether to overwrite the existing astrometry solution.
        verbose : bool, optional
            Whether to print verbose output.
            
        Returns
        -------
        output_img : ScienceImage
            The output image with astrometry solution.
        """
        from ezphot.methods import Platesolve
        platesolve = Platesolve()
        self = platesolve.solve_astrometry(
            target_img = self,
            overwrite = overwrite,
            verbose = verbose
            )
        
        self = platesolve.solve_scamp(
            target_img = self,
            scamp_sexparams = scamp_sexparams,
            scamp_params = scamp_params,
            overwrite = overwrite,
            verbose = verbose
        )
        
        return self
        
    def calculate_invalidmask(self,
                              threshold_invalid_connection: int = 100000,
                              save: bool = False,
                              verbose: bool = True,
                              visualize: bool = True,
                              save_fig: bool = False
                              ):
        """ 
        Calculate the invalid mask for this ScienceImage.
        The invalid mask is a mask of pixels that are invalid (zero or nan value).
        If save is True, the invalid mask is saved. Then, you can load the invalid mask with `self.invalidmask`.
        
        Parameters
        ----------
        threshold_invalid_connection : int
            The threshold for invalid connection.
        save : bool
            If True, save the invalid mask.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the invalid mask.
        save_fig : bool
            If True, save the figure of the invalid mask.
            
        Returns
        -------
        target_ivpmask : Mask
            The invalid mask.
        """
        from ezphot.methods import MaskGenerator
        maskgenerator = MaskGenerator()
        target_ivpmask = maskgenerator.mask_invalidpixel(
            target_img = self,
            threshold_invalid_connection= threshold_invalid_connection,
            # Others
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig
            )
        return target_ivpmask
    
    def calculate_circularmask(self,
                               target_srcmask: Mask = None,
                               x_position: float = None,
                               y_position: float = None,
                               radius_arcsec: float = None,
                               unit = 'coord',
                               save: bool = False,
                               verbose: bool = True,
                               visualize: bool = True,
                               save_fig: bool = False
                               ):
        """
        Calculate the circular mask for this ScienceImage.
        The circular mask is a mask of pixels that are within a circular region.
        If save is True, the circular mask is saved. Then, you can load the circular mask with `self.circularmask`.
        
        Parameters
        ----------
        target_srcmask : Mask
            The source mask.
        x_position : float
            The x position of the center of the circular mask.
        y_position : float
            The y position of the center of the circular mask.
        radius_arcsec : float
            The radius of the circular mask in arcseconds.
        unit : str
            The unit of the x and y position. 'coord' for coordinate, 'pixel' for pixel.
        save : bool 
            If True, save the circular mask.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the circular mask.
        save_fig : bool
            If True, save the figure of the circular mask.
        
        Returns 
        -------
        target_circularmask : Mask
            The circular mask.
        """
        from ezphot.methods import MaskGenerator
        maskgenerator = MaskGenerator()
        target_sourcemask = maskgenerator.mask_circle(
            target_img = self,
            target_mask = target_srcmask,
            mask_type = 'source',
            x_position = x_position,
            y_position = y_position,
            radius_arcsec = radius_arcsec,
            unit = unit,
            
            # Others
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig
            )
        return target_sourcemask
        
    def calculate_sourcemask(self,
                             target_srcmask: Mask = None,
                             sigma: float = 5.0,
                             mask_radius_factor: float = 3,
                             saturation_level: float = 50000,
                             save: bool = False,
                             verbose: bool = True,
                             visualize: bool = True,
                             save_fig: bool = False
                             ):
        """ 
        Calculate the source mask for this ScienceImage.
        The source mask is a mask of pixels that are sources. Detection is made with global background and background RMS map.
        If save is True, the source mask is saved. Then, you can load the source mask with `self.sourcemask`.
        
        Parameters
        ----------  
        target_srcmask : Mask
            The source mask. 
        sigma : float
            The sigma for the source detection.
        mask_radius_factor : float
            The radius factor for the source detection.
        saturation_level : float
            The saturation level for the source detection.
        save : bool
            If True, save the source mask.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the source mask.
        save_fig : bool
            If True, save the figure of the source mask.
            
        Returns
        -------
        target_sourcemask : Mask
            The source mask.
        """
        from ezphot.methods import MaskGenerator
        maskgenerator = MaskGenerator()
        target_sourcemask = maskgenerator.mask_sources(
            target_img = self,
            target_mask = target_srcmask,
            sigma = sigma,
            mask_radius_factor = mask_radius_factor,
            saturation_level = saturation_level,

            # Others
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig
            )
        return target_sourcemask

    def calculate_bkg(self,
                      target_srcmask: Mask = None,
                      target_ivpmask: Mask = None,
                      is_2D_bkg: bool = True,
                      box_size: int = 64,
                      filter_size: int = 3,
                      correct_global_offset: bool = True,
                      save: bool = False,
                      verbose: bool = True,
                      visualize: bool = True,
                      save_fig: bool = False):
        """
        Calculate the background map for this ScienceImage.
        The background map is a map of the background level of the image.
        If save is True, the background map is saved. Then, you can load the background map with `self.bkgmap`.
        
        Parameters
        ----------
        target_srcmask : Mask
            The source mask.
        target_ivpmask : Mask
            The invalid mask.
        is_2D_bkg : bool
            If True, use 2D background estimation.
        box_size : int
            The box size for the background estimation.
        filter_size : int
            The filter size for the background estimation.
        correct_global_offset : bool
            If True, correct the global offset of the background map.
        save : bool
            If True, save the background map.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the background map.
        save_fig : bool
            If True, save the figure of the background map.
        
        Returns
        -------
        target_bkg : Background
            The calculated background map.
        """
        from ezphot.methods import BackgroundGenerator
        bkggenerator = BackgroundGenerator()
        target_bkg, _ = bkggenerator.estimate_with_sep(
            target_img = self,
            target_srcmask = target_srcmask,
            target_ivpmask = target_ivpmask,
            is_2D_bkg = is_2D_bkg,
            box_size = box_size,
            filter_size = filter_size,
            correct_global_offset = correct_global_offset,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)

        return target_bkg
    
    def calculate_bkgrms(self, 
                         target_srcmask: Mask = None,
                         target_ivpmask: Mask = None,
                         box_size: int = 64,
                         filter_size: int = 3,
                         save: bool = False,
                         verbose: bool = True,
                         visualize: bool = True,
                         save_fig: bool = False):
        """
        Calculate the background RMS map for this ScienceImage.
        The background RMS map is a map of the background RMS level of the image.
        If save is True, the background RMS map is saved. Then, you can load the background RMS map with `self.bkgrms`.
        
        Parameters
        ----------
        target_srcmask : Mask
            The source mask.
        target_ivpmask : Mask
            The invalid mask.
        box_size : int
            The box size for the background RMS estimation.
        filter_size : int
            The filter size for the background RMS estimation.
        save : bool
            If True, save the background RMS map.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the background RMS map.
        save_fig : bool
            If True, save the figure of the background RMS map.
        
        Returns
        -------
        target_bkgrms : Errormap
            The background RMS map.
        """
        from ezphot.methods import ErrormapGenerator
        errormapgenerator = ErrormapGenerator()
        target_bkgrms, _, _ = errormapgenerator.calculate_errormap_from_image(
            target_img = self,  
            target_srcmask = target_srcmask,
            target_ivpmask = target_ivpmask,
            box_size = box_size,
            filter_size = filter_size,
            erormap_type = 'bkgrms',
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
        return target_bkgrms
    
    def calculate_bkgrms_from_propagation(self,
                                          target_bkg: Background = None,
                                          mbias: CalibrationImage = None,
                                          mdark: CalibrationImage = None,
                                          mflat: CalibrationImage = None,
                                          mflaterr: Errormap = None,
                                          save: bool = False,
                                          verbose: bool = True,
                                          visualize: bool = True,
                                          save_fig: bool = False):
        """
        Calculate the background RMS map for this ScienceImage from the background map, bias frame, dark frame, and flat frame.
        The background RMS map is a map of the background RMS level of the image. 
        If save is True, the background RMS map is saved. Then, you can load the background RMS map with `self.bkgrms`.

        Parameters
        ----------
        target_srcmask : Mask
            The source mask.
        target_ivpmask : Mask
            The invalid mask.
        mbias : CalibrationImage
            The bias frame.
        mdark : CalibrationImage
            The dark frame.
        mflat : CalibrationImage
            The flat frame.
        mflaterr : Errormap
            The flat error map.
        save : bool
            If True, save the background RMS map.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the background RMS map.
        save_fig : bool
            If True, save the figure of the background RMS map.

        Returns
        -------
        target_bkgrms : Errormap
            The background RMS map.
        """
        
        from ezphot.methods import Preprocess
        preprocess = Preprocess()
        # prepare the data
        if target_bkg is None:
            if self.bkgmap is None:
                raise ValueError("Cannot calculate background RMS map: Input background map. OR Register background map with scienceimage.calculate_background(save = True) first.")
            else:
                target_bkg = self.bkgmap
        
        mbias_path, mdark_path, mflat_path = None, None, None
        if mbias is None:
            mbias = preprocess.get_masterframe_from_image(self, 'bias', 30)[0]
        if mdark is None:
            mdark = preprocess.get_masterframe_from_image(self, 'dark', 30)[0]
        if mflat is None:
            mflat = preprocess.get_masterframe_from_image(self, 'flat', 30)[0]
        if mbias is None or mdark is None or mflat is None:
            raise ValueError("Cannot calculate background RMS: required calibration frames are missing.")

        from ezphot.methods import ErrormapGenerator
        errormapgenerator = ErrormapGenerator()
        target_bkgrms = errormapgenerator.calculate_bkgrms_from_propagation(
            target_img = self,
            target_bkg = target_bkg,
            mbias_img = mbias,
            mdark_img = mdark,
            mflat_img = mflat,
            mflaterr_img = mflaterr,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
                
        return target_bkgrms
        
    def calculate_errormap(self, 
                           target_srcmask: Mask = None,
                           target_ivpmask: Mask = None,
                           box_size: int = 64,
                           filter_size: int = 3,
                           save: bool = False,
                           verbose: bool = True,
                           visualize: bool = True,
                           save_fig: bool = False):
        """
        Calculate the source RMS map for this ScienceImage.
        The source RMS map is a map of the source RMS level of the image.
        If save is True, the source RMS map is saved. Then, you can load the source RMS map with `self.sourcerms`.

        Parameters
        ----------
        target_srcmask : Mask
            The source mask.
        target_ivpmask : Mask
            The invalid mask.
        box_size : int
            The box size for the source RMS estimation.
        filter_size : int
            The filter size for the source RMS estimation.
        save : bool
            If True, save the source RMS map.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the source RMS map.
        save_fig : bool
            If True, save the figure of the source RMS map.

        Returns
        -------
        target_sourcerms : Errormap
            The source RMS map.
        """
        from ezphot.methods import ErrormapGenerator
        errormapgenerator = ErrormapGenerator()
        target_sourcerms = errormapgenerator.calculate_errormap_from_image(
            target_img = self,  
            target_srcmask = target_srcmask,
            target_ivpmask = target_ivpmask,
            box_size = box_size,
            filter_size = filter_size,
            erormap_type = 'sourcerms',
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
        return target_sourcerms
    
    def calculate_errormap_from_propagation(self,
                                            mbias: CalibrationImage = None,
                                            mdark: CalibrationImage = None,
                                            mflat: CalibrationImage = None,
                                            mflaterr: Errormap = None,
                                            save: bool = False,
                                            verbose: bool = True,
                                            visualize: bool = True,
                                            save_fig: bool = False):
        """
        Calculate the source RMS map for this ScienceImage from the background map, bias frame, dark frame, and flat frame.
        The source RMS map is a map of the source RMS level of the image.
        If save is True, the source RMS map is saved. Then, you can load the source RMS map with `self.sourcerms`.

        Parameters
        ----------
        target_srcmask : Mask
            The source mask.
        target_ivpmask : Mask
            The invalid mask.
        mbias : CalibrationImage
            The bias frame.
        mdark : CalibrationImage
            The dark frame.
        mflat : CalibrationImage
            The flat frame.
        mflaterr : Errormap
            The flat error map.
        save : bool
            If True, save the source RMS map.
        verbose : bool
            If True, print verbose output.
        visualize : bool
            If True, visualize the source RMS map.
        save_fig : bool
            If True, save the figure of the source RMS map.

        Returns
        -------
        target_sourcerms : Errormap
            The source RMS map.
        """
        from ezphot.methods import Preprocess
        preprocess = Preprocess()
        # prepare the data

        mbias_path, mdark_path, mflat_path = None, None, None
        if mbias is None:
            mbias = preprocess.get_masterframe_from_image(self, 'bias', 30)[0]
        if mdark is None:
            mdark = preprocess.get_masterframe_from_image(self, 'dark', 30)[0]
        if mflat is None:
            mflat = preprocess.get_masterframe_from_image(self, 'flat', 30)[0]
        if mbias is None or mdark is None or mflat is None:
            raise ValueError("Cannot calculate source RMS: required calibration frames are missing.")

        from ezphot.methods import ErrormapGenerator
        errormapgenerator = ErrormapGenerator()
        target_sourcerms = errormapgenerator.calculate_sourcerms_from_propagation(
            target_img = self,
            mbias_img = mbias,
            mdark_img = mdark,
            mflat_img = mflat,
            mflaterr_img = mflaterr,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
            
        return target_sourcerms
    
    def get_referenceframe(self, 
                           telname: str = None,
                           min_obsdate: Union[str, float, Time] = None,
                           max_obsdate: Union[str, float, Time] = None,
                           sort_key: Union[str, List[str]] = ['fraction', 'depth'],
                           overlap_threshold: float = 0.5,
                           return_groups: bool = True,
                           group_overlap_threshold: float = 0.8,
                           verbose: bool = True
                           ):
        """
        Get the reference frame from the target image.
        
        Parameters
        ----------
        telname : str, optional
            The telescope name.
        min_obsdate : Union[str, float, Time], optional
            The minimum observation date.
        max_obsdate : Union[str, float, Time], optional
            The maximum observation date.
        sort_key : Union[str, List[str]], optional
            The sort key.
        overlap_threshold : float, optional
            The overlap threshold.
        return_groups : bool, optional
            Whether to return the groups.
        group_overlap_threshold : float, optional
            The group overlap threshold.
            
        Returns
        -------
        reference_img : ReferenceImage
            The reference image.
        reference_frames : Table
            The metadata of the reference frames matched the criteria.
        """
        from ezphot.methods import Subtract
        subtract = Subtract()
        result = subtract.get_referenceframe_from_image(
            target_img = self,
            telname = telname,
            min_obsdate = min_obsdate,
            max_obsdate = max_obsdate,
            sort_key = sort_key,
            overlap_threshold = overlap_threshold,
            return_groups = return_groups,
            group_overlap_threshold = group_overlap_threshold,
            verbose = verbose)
        return result
    
    def get_masterframe(self,
                        imagetyp: str,
                        max_days: float = 10
                        ):
        """
        Get master frame from the image.
        
        This method will search for the master frame in the master frame directory.
        
        Parameters
        ----------
        imagetyp : str
            The type of image to get the master frame from. (BIAS, DARK, FLAT)
        max_days : float, optional
            The maximum number of days to search for the master frame.
            
        Returns
        -------
        master_img : CalibrationImage
            The master frame image.
        master_frames_tbl : Table
            Metadata of the master frame(s) found.
        """
        from ezphot.methods import Preprocess
        preprocess = Preprocess()
        result = preprocess.get_masterframe_from_image(
            target_img = self,
            imagetyp = imagetyp,
            max_days = max_days,
            )
                
        return result
    
    def get_referencecatalog(self,
                             catalog_type: str = 'GAIAXP',
                             fraction_criteria: float = 0.05,
                             query_when_not_archived: bool = True,
                             verbose: bool = False):
        from ezphot.skycatalog import SkyCatalogUtility
        skycatalogutility = SkyCatalogUtility()
        skycatalog = skycatalogutility.get_catalogs(
            ra = self.ra, 
            dec = self.dec, 
            objname = self.objname, 
            fov_ra = self.fov_ra, 
            fov_dec = self.fov_dec, 
            catalog_type = catalog_type, 
            fraction_criteria = fraction_criteria, 
            query_when_not_archived = query_when_not_archived, 
            verbose = verbose)
        return skycatalog
    
    def query_referenceframe(self,
                             save_path: str = None,
                             verbose: bool = True,
                             n_processes: int = 4):
        """
        Query the reference frame from the target image.
        
        Parameters
        ----------
        save_path : str, optional
            The save path of the reference frame.
        verbose : bool, optional
            The verbose flag.
        n_processes : int, optional
            The number of processes.
        
        Returns
        -------
        reference_img : ReferenceImage
            The reference image.
        """
        from ezphot.utils import ImageQuerier
        imagequerier = ImageQuerier()
        result = imagequerier.query(
            width = self.naxis1,
            height = self.naxis2,
            ra = self.center['ra'],
            dec = self.center['dec'],
            pixelscale = self.pixelscale[0],
            telinfo = self.telinfo,
            save_path = save_path,
            objname = self.objname,
            rotation_angle = 0.0,
            verbose = verbose,
            n_processes = n_processes)
        return result
    
    def subtract_background(self, 
                            target_bkg: Background,
                            overwrite: bool = False,
                            save: bool = False,
                            verbose: bool = True,
                            visualize: bool = True,
                            save_fig: bool = False):
        """
        Subtract the background from the target image.
        
        Parameters
        ----------
        target_bkg : ezphot.imageobjects.Background
            The background image to subtract from the target image.
        overwrite : bool, optional
            Whether to overwrite the target image.
        save : bool, optional
            Whether to save the subtracted image.
        verbose : bool, optional
            Whether to print verbose output.
        visualize : bool, optional
            Whether to visualize the subtracted image.
        save_fig : bool, optional
            Whether to save the subtracted image as a figure.
            
        Returns
        -------
        target_img : ScienceImage or ReferenceImage
            The subtracted image.
        """
        
        from ezphot.methods import BackgroundGenerator
        backgroundgenerator = BackgroundGenerator()
        result = backgroundgenerator.subtract_background(
            target_img = self,
            target_bkg = target_bkg,
            overwrite = overwrite,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
        return result
    
    def photometry_sex(self,
                       target_bkg: Background = None,
                       target_bkgrms: Errormap = None,
                       target_mask: Mask = None,
                       sex_params: dict = None,
                       detection_sigma: float = 5,
                       aperture_diameter_arcsec: Union[float, list] = [5,7,10],
                       aperture_diameter_seeing: Union[float, list] = [3.5,4.5],
                       annulus_width_arcsec: Union[float, list] = None,
                       saturation_level: float = 60000,
                       kron_factor: float = 2.5,
                       save: bool = True,
                       verbose: bool = True,
                       visualize: bool = True,
                       save_fig: bool = False):
        """
        Perform source-extractor photometry on the target image.
        
        Parameters
        ----------
        target_bkg : Background
            The background to subtract from the target image.
        target_bkgrms : Errormap
            The background RMS to use for the photometry.
        target_mask : Mask
            The mask to use for the photometry.
        sex_params : dict
            The source-extractor parameters.
        detection_sigma : float
            The detection sigma for the photometry.
        aperture_diameter_arcsec : Union[float, list]
            The aperture diameter in arcseconds.
        aperture_diameter_seeing : Union[float, list]
            The aperture diameter in seeing units.
        saturation_level : float
            The saturation level for the photometry.
        kron_factor : float
            The Kron factor for the photometry.
        save : bool
            Whether to save the photometry catalog.
        verbose : bool
            Whether to print verbose output.
        visualize : bool
            Whether to visualize the photometry.
        save_fig : bool
            Whether to save the photometry figure.

        Returns
        -------
        result : Catalog
            The photometry catalog.
        """
        
        from ezphot.methods import AperturePhotometry
        aperturephotometry = AperturePhotometry()
        result = aperturephotometry.sex_photometry(
            target_img = self,
            target_bkg = target_bkg,
            target_bkgrms = target_bkgrms,
            target_mask = target_mask,
            sex_params = sex_params,
            detection_sigma = detection_sigma,
            aperture_diameter_arcsec = aperture_diameter_arcsec,
            aperture_diameter_seeing = aperture_diameter_seeing,
            annulus_width_arcsec = annulus_width_arcsec,
            saturation_level = saturation_level,
            kron_factor = kron_factor,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
        return result
    
    def photometry_forced_circular(self,
                                   x_arr: list = None,
                                   y_arr: list = None,
                                   aperture_diameter_arcsec: Union[float, list] = [5,7,10],
                                   aperture_diameter_seeing: Union[float, list] = [3.5,4.5],
                                   annulus_width_arcsec: Union[float, list] = None,
                                   unit: str = 'coord',
                                   target_bkg: Background = None,
                                   target_bkgrms: Errormap = None,
                                   target_mask: Mask = None,
                                   
                                   save: bool = True,
                                   verbose: bool = True,
                                   visualize: bool = True,
                                   save_fig: bool = False):
        '''
        Perform forced photometry on the target image.
        
        Parameters
        ----------
        x_arr : list
            The x-coordinates of the sources.
        y_arr : list
            The y-coordinates of the sources.
        aperture_diameter_arcsec : Union[float, list]
            The aperture diameter in arcseconds.
        aperture_diameter_seeing : Union[float, list]
            The aperture diameter in seeing units.
        annulus_width_arcsec : Union[float, list]
            The annulus width in arcseconds.
        unit : str
            The unit of the coordinates.
        target_bkg : Background
            The background to subtract from the target image.
        target_bkgrms : Errormap
            The background RMS to use for the photometry.
        target_mask : Mask
            The mask to use for the photometry.
        save : bool
            Whether to save the photometry catalog.
        verbose : bool
            Whether to print verbose output.
        visualize : bool
            Whether to visualize the photometry.
        save_fig : bool
            Whether to save the photometry figure.

        Returns
        -------
        result : Catalog
            The photometry catalog.
        '''

        from ezphot.methods import AperturePhotometry
        aperturephotometry = AperturePhotometry()
        result = aperturephotometry.circular_photometry(
            target_img = self,
            target_bkg = target_bkg,
            target_bkgrms = target_bkgrms,
            target_mask = target_mask,
            x_arr = x_arr,
            y_arr = y_arr,
            aperture_diameter_arcsec = aperture_diameter_arcsec,
            aperture_diameter_seeing = aperture_diameter_seeing,
            annulus_width_arcsec = annulus_width_arcsec,
            unit = unit,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
        return result
        
    def photometry_forced_elliptical(self,
                                     x_arr: list = None,
                                     y_arr: list = None,
                                     sma_arr: list = None,
                                     smi_arr: list = None,
                                     theta_arr: list = None,
                                     unit: str = 'pixel',
                                     annulus_ratio: float = None,
                                     target_bkg: Background = None,
                                     target_bkgrms: Errormap = None,
                                     target_mask: Mask = None,
                                     
                                     save: bool = True,
                                     verbose: bool = True,
                                     visualize: bool = True,
                                     save_fig: bool = False):
        '''
        Perform forced photometry on the target image.
        
        Parameters
        ----------
        x_arr : list
            The x-coordinates of the sources.
        y_arr : list
            The y-coordinates of the sources.
        sma_arr : list
            The semi-major axis of the sources.
        smi_arr : list
            The semi-minor axis of the sources.
        theta_arr : list
            The position angle of the sources.
        unit : str
            The unit of the coordinates.
        annulus_ratio : float
            The ratio of the annulus diameter to the semi-major axis.
        target_bkg : Background
            The background to subtract from the target image.
        target_bkgrms : Errormap
            The background RMS to use for the photometry.
        target_mask : Mask
            The mask to use for the photometry.
        save : bool
            Whether to save the photometry catalog.
        verbose : bool
            Whether to print verbose output.
        visualize : bool
            Whether to visualize the photometry.
        save_fig : bool
            Whether to save the photometry figure.

        Returns
        -------
        result : Catalog
            The photometry catalog.
        '''
        
        from ezphot.methods import AperturePhotometry
        aperturephotometry = AperturePhotometry()
        result = aperturephotometry.elliptical_photometry(
            target_img = self,
            target_bkg = target_bkg,
            target_bkgrms = target_bkgrms,
            target_mask = target_mask,
            x_arr = x_arr,
            y_arr = y_arr,
            sma_arr = sma_arr,
            smi_arr = smi_arr,
            theta_arr = theta_arr,
            unit = unit,
            annulus_ratio = annulus_ratio,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig)
        return result
        
    def photometric_calibration(self, 
                                target_catalog,
                                catalog_type: str = 'GAIAXP',
                                catalog_version: str = 'v1',
                                max_distance_second: float = 2.5,
                                min_number_of_sources: int = 10,
                                calculate_color_terms: bool = True,
                                calculate_mag_terms: bool = True,
                                
                                # Selection parameters
                                mag_lower: float = 12,
                                mag_upper: float = 16,
                                snr_lower: float = 10,
                                snr_upper: float = 150,
                                dynamic_mag_range: bool = False,
                                classstar_lower: float = 0.8,
                                elongation_upper: float = 1.7,
                                elongation_sigma: float = 5,
                                fwhm_lower_arcsec: float = 1,
                                fwhm_upper_arcsec: float = 10,
                                fwhm_sigma: float = 5,
                                flag_upper: int = 1,
                                maskflag_upper: int = 1,
                                ra_deg: float = None,
                                dec_deg: float = None,
                                radius_arcsec: float = 1500,  
                                inner_fraction: float = 0.7, # Fraction of the images
                                isolation_radius_arcsec: float = 10.0,
                                
                                magnitude_key: str = 'MAG_AUTO',
                                magnitudeerr_key: str = 'MAGERR_AUTO',
                                fwhm_key: str = 'FWHM_WORLD',
                                ra_key: str = 'X_WORLD',
                                dec_key: str = 'Y_WORLD',
                                classstar_key: str = 'CLASS_STAR',
                                elongation_key: str = 'ELONGATION',
                                flag_key: str = 'FLAGS',
                                maskflag_key: str = 'IMAFLAGS_ISO',

                                # Other parameters
                                update_header: bool = True,
                                save: bool = True,
                                verbose: bool = True,
                                visualize: bool = True,
                                save_fig: bool = False,
                                save_refcat: bool = True
                                ):
        
        from ezphot.methods import PhotometricCalibration
        from ezphot.dataobjects import Catalog
        if not isinstance(target_catalog, Catalog):
            raise ValueError("target_catalog must be a Catalog object.")
        photometriccalibration = PhotometricCalibration()
    
        result = photometriccalibration.photometric_calibration(
            target_img = self,
            target_catalog = target_catalog,
            catalog_type = catalog_type,
            catalog_version = catalog_version,
            max_distance_second = max_distance_second,
            min_number_of_sources = min_number_of_sources,
            calculate_color_terms = calculate_color_terms,
            calculate_mag_terms = calculate_mag_terms,
            mag_lower = mag_lower,
            mag_upper = mag_upper,
            snr_lower = snr_lower,
            snr_upper = snr_upper,
            dynamic_mag_range = dynamic_mag_range,
            classstar_lower = classstar_lower,
            elongation_upper = elongation_upper,
            elongation_sigma = elongation_sigma,
            fwhm_lower_arcsec = fwhm_lower_arcsec,
            fwhm_upper_arcsec = fwhm_upper_arcsec,
            fwhm_sigma = fwhm_sigma,
            flag_upper = flag_upper,
            maskflag_upper = maskflag_upper,
            ra_deg = ra_deg,
            dec_deg = dec_deg,
            radius_arcsec = radius_arcsec,
            inner_fraction = inner_fraction,
            isolation_radius_arcsec = isolation_radius_arcsec,
            magnitude_key = magnitude_key,
            magnitudeerr_key = magnitudeerr_key,
            fwhm_key = fwhm_key,
            ra_key = ra_key,
            dec_key = dec_key,
            classstar_key = classstar_key,
            elongation_key = elongation_key,
            flag_key = flag_key,
            maskflag_key = maskflag_key,
            update_header = update_header,
            save = save,
            verbose = verbose,
            visualize = visualize,
            save_fig = save_fig,
            save_refcat = save_refcat)
        
        return result
    
    def apply_color_terms(self,
                          target_catalog,
                          comparison_catalog = None,
                          magkey_suffix: Union[str, List[str]] = 'MAGSKY_',
                          max_distance_arcsec: float = 2.5,
                          verbose: bool = True,
                          save: bool = True):
        from ezphot.methods import PhotometricCalibration
        from ezphot.dataobjects import Catalog
        if not isinstance(target_catalog, Catalog):
            raise ValueError("target_catalog must be a Catalog object.")
        if comparison_catalog is not None and not isinstance(comparison_catalog, Catalog):
            raise ValueError("comparison_catalog must be a Catalog object.")
        photometriccalibration = PhotometricCalibration()
        target_catalog = photometriccalibration.apply_color_terms(target_img = self, 
                                                                  target_catalog = target_catalog, 
                                                                  comparison_catalog = comparison_catalog,
                                                                  magkey_suffix = magkey_suffix,
                                                                  max_distance_arcsec = max_distance_arcsec,
                                                                  verbose = verbose, 
                                                                  save = save)
        return target_catalog
    
    def apply_mag_terms(self,
                        target_catalog,
                        magkey_suffix: Union[str, List[str]] = 'MAGSKY_',
                        verbose: bool = True,
                        save: bool = True):
        from ezphot.methods import PhotometricCalibration
        from ezphot.dataobjects import Catalog
        if not isinstance(target_catalog, Catalog):
            raise ValueError("target_catalog must be a Catalog object.")
        photometriccalibration = PhotometricCalibration()
        target_catalog = photometriccalibration.apply_mag_terms(target_img = self, 
                                                                target_catalog = target_catalog, 
                                                                verbose = verbose, 
                                                                save = save)
        return target_catalog
    
    def apply_zp(self,
                 target_catalog,
                 verbose: bool = True,
                 save: bool = True):
        from ezphot.methods import PhotometricCalibration
        from ezphot.dataobjects import Catalog
        if not isinstance(target_catalog, Catalog):
            raise ValueError("target_catalog must be a Catalog object.")
        photometriccalibration = PhotometricCalibration()
        target_catalog = photometriccalibration.apply_zp(target_img = self, 
                                                         target_catalog = target_catalog,
                                                         verbose = verbose, 
                                                         save = save)
        return target_catalog

    def reproject(self, 
                  target_errormap: Optional[Errormap] = None,
                  swarp_params: Optional[dict] = None,
                  resample_type: str = 'LANCZOS3',
                  center_ra: Optional[float] = None,
                  center_dec: Optional[float] = None,
                  x_size: Optional[int] = None,
                  y_size: Optional[int] = None,
                  pixelscale: Optional[float] = None,
                  verbose: bool = True,
                  overwrite: bool = False,
                  save: bool = True,
                  return_ivpmask: bool = False,
                  fill_zero_tonan: bool = True):
        """
        Reproject the image with SWarp.
        
        Parameters
        ----------
        target_errormap : Errormap
            The error map to use for the reproject.
        swarp_params : dict
            The parameters for SWarp.
        resample_type : str
            The type of resampling for SWarp.
        center_ra : float
            The center RA for SWarp.
        center_dec : float
            The center Dec for SWarp.
        x_size : int
            The size of the image in the x direction for SWarp.
        y_size : int
            The size of the image in the y direction for SWarp.
        pixelscale : float
            The pixel scale for SWarp.
        verbose : bool
            Whether to print verbose output.
        overwrite : bool
            Whether to overwrite the existing reprojected image.
        save : bool
            Whether to save the reprojected image.
        return_ivpmask : bool
            Whether to return the invalid pixel mask.
        fill_zero_tonan : bool
            Whether to fill the zero to nan.
        """
        from ezphot.methods import Reproject
        reproject = Reproject()
        result = reproject.reproject(
            target_img = self,
            target_errormap = target_errormap,
            swarp_params = swarp_params,
            resample_type = resample_type,
            center_ra = center_ra,
            center_dec = center_dec,
            x_size = x_size,
            y_size = y_size,
            pixelscale = pixelscale,
            verbose = verbose,
            overwrite = overwrite,
            save = save,
            return_ivpmask = return_ivpmask,
            fill_zero_tonan = fill_zero_tonan)
        return result
    
    def DIA(self,
            reference_img: "ReferenceImage" = None,
            target_bkg: Background = None,
            
            # Photometry configuration
            detection_sigma: float = 1.5,
            aperture_diameter_arcsec: List[float] = [5, 7, 10],
            aperture_diameter_seeing: List[float] = [3.5, 4.5],
            kron_factor: float = 1.5,
            catalog_type: str = 'GAIAXP',
            
            # DIA configuration
            reject_variable_sources: bool = False,
            negative_detection: bool = True,
            reverse_subtraction: bool = False,            
            save_transient_figure: bool = True,
            save_candidate_figure: bool = True,
            show_transient_numbers: int = 100,
            show_candidate_numbers: int = 100,
            # HOTPANTS parameters
            convim: str = 't',
            normim: str = 'i',
            tu: float = None,
            tl: float = None,
            iu: float = None,
            il: float = None,
            nrx: int = 1,
            nry: int = 1,
            nsx: int = 10,
            nsy: int = 10,
            ko: int = 2,
            bgo: int = 1,
            r: int = 10,
            
            # Other parameters
            save: bool = True,
            verbose: bool = True,
            visualize: bool = False
            ):
        
        from ezphot.methods import Subtract
        if reference_img is None:
            reference_result = self.get_referenceframe(verbose = verbose)
            if reference_result is None:
                raise RuntimeError("No reference image found.")            
            reference_img = reference_result[0]
        
        reference_img = np.atleast_1d(reference_img)
        
        subtract = Subtract()
        result = subtract.find_transients(
            target_img = self,
            reference_imglist = reference_img,
            target_bkg = target_bkg,
            detection_sigma = detection_sigma,
            aperture_diameter_arcsec = aperture_diameter_arcsec,
            aperture_diameter_seeing = aperture_diameter_seeing,
            kron_factor = kron_factor,
            catalog_type = catalog_type,
            reject_variable_sources = reject_variable_sources,
            negative_detection = negative_detection,
            reverse_subtraction = reverse_subtraction,
            save_transient_figure = save_transient_figure,
            save_candidate_figure = save_candidate_figure,
            show_transient_numbers = show_transient_numbers,
            show_candidate_numbers = show_candidate_numbers,
            convim = convim,
            normim = normim,
            tu = tu,
            tl = tl,
            iu = iu,
            il = il,
            nrx = nrx,
            nry = nry,
            nsx = nsx,
            nsy = nsy,
            ko = ko,
            bgo = bgo,
            r = r,
            save = save,
            verbose = verbose,
            visualize = visualize,
        )
        return result



#%%
import inspect
from typing import Union, Optional, List

from ezphot.error import PlatesolveError
from ezphot.helper import Helper
from ezphot.imageobjects import ScienceImage, ReferenceImage
#%%
class Platesolve:
    """
    Platesolve class for solving astrometry and SCAMP.
    
    This class provides methods 
    
    1. Solving astrometry using Astrometry.net.
    
    2. Solving astrometry using SCAMP.
    """
    
    def __init__(self):
        """
        Initialize the Platesolve class.
        
        Parameters
        ----------
        None
        
        Returns
        -------
        None
        """
        self.helper = Helper()

    def __repr__(self):
        return f"Method class: {self.__class__.__name__}\n For help, use 'help(self)' or `self.help()`."

    def help(self):
        # Get all public methods from the class, excluding `help`
        methods = [
            (name, obj)
            for name, obj in inspect.getmembers(self.__class__, inspect.isfunction)
            if not name.startswith("_") and name != "help"
        ]

        # Build plain text list with parameters
        lines = []
        for name, func in methods:
            sig = inspect.signature(func)
            params = [str(p) for p in sig.parameters.values() if p.name != "self"]
            sig_str = f"({', '.join(params)})" if params else "()"
            lines.append(f"- {name}{sig_str}")

        # Final plain text output
        help_text = ""
        self.helper.print(f"Help for {self.__class__.__name__}\n{help_text}\nPublic methods:\n" + "\n".join(lines), True)

    def solve_astrometry(self,
                         # Input parameters
                         target_img: Union[ScienceImage, ReferenceImage],
                        
                         # Other parameters
                         overwrite: bool = True,
                         verbose: bool = True,
                         **kwargs
                         ):
        """
        Solve astrometry using Astrometry.net.
        
        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage]
            The target image to solve astrometry for.
        overwrite : bool, optional
            Whether to overwrite the existing astrometry solution.
        verbose : bool, optional
            Whether to print verbose output.
            
        Returns
        -------
        output_img : ScienceImage
            The output image with astrometry solution.
        """
        if overwrite:
            target_outpath = target_img.savepath.savepath
        else:
            target_outpath = target_img.savepath.savepath.parent / f'astrometry_{target_img.savepath.savepath.name}'
            
        result, astrometry_output_images = self.helper.run_astrometry(
            target_path = target_img.path,
            astrometry_sexconfigfile = target_img.config['ASTROMETRY_SEXCONFIG'],
            ra = target_img.ra,
            dec = target_img.dec,
            radius = max(target_img.fovx, target_img.fovy) / 2,
            pixelscale = target_img.telinfo['pixelscale'],
            target_outpath = target_outpath,
            verbose = verbose,
        )
        if not result:
            raise PlatesolveError("Astrometry failed", target_img.path)
        else:
            output_img = type(target_img)(path = astrometry_output_images, telinfo = target_img.telinfo, status = target_img.status.copy(), load = True)
            output_img.update_status('ASTROMETRY')
            return output_img
        
    def solve_scamp(self,
                    # Input parameters
                    target_img: Optional[Union[ScienceImage, ReferenceImage, List[ScienceImage], List[ReferenceImage]]],
                    scamp_sexparams: dict = None,
                    scamp_params: dict = None,
                    # Other parameters
                    overwrite: bool = True,
                    verbose: bool = True,
                    ):
        """
        Solve astrometry using SCAMP.
        
        Parameters
        ----------
        target_img : Union[ScienceImage, ReferenceImage, List[ScienceImage], List[ReferenceImage]]
            The target image(s) to solve astrometry for.
        scamp_sexparams : dict, optional
            The SExtractor parameters for SCAMP.
        scamp_params : dict, optional
            The SCAMP parameters for SCAMP.
        overwrite : bool, optional
            Whether to overwrite the existing astrometry solution.
        verbose : bool, optional
            Whether to print verbose output.
            
        Returns 
        -------
        output_imglist : List[ScienceImage]
            The output image(s) with astrometry solution.
        """
        target_imglist = target_img if isinstance(target_img, list) else [target_img]
        target_imglist_path = [target_img.path for target_img in target_imglist]
        
        if overwrite:
            output_dir = target_imglist[0].savepath.savedir
        else:
            output_dir = target_imglist[0].path.parent
            
        scamp_results, scamp_output_images = self.helper.run_scamp(
            target_path = target_imglist_path,
            scamp_sexconfigfile = target_imglist[0].config['SCAMP_SEXCONFIG'],
            scamp_configfile = target_imglist[0].config['SCAMP_CONFIG'],
            scamp_sexparams = scamp_sexparams,
            scamp_params = scamp_params,
            output_dir = output_dir,
            overwrite = overwrite,
            verbose = verbose,            
        )
        
        if not all(scamp_results):
            raise PlatesolveError(f"SCAMP failed for {target_imglist_path}")
        else:    
            output_imglist = []
            for target_img, output_path in zip(target_imglist, scamp_output_images):
                output_img = type(target_imglist[0])(path = output_path, telinfo = target_img.telinfo, status = target_img.status.copy(), load = True)
                output_img.update_status('SCAMP')
                output_imglist.append(output_img)
            return output_imglist
    
# %%
if __name__ == '__main__':
    target_img = ScienceImage(path = '/home/hhchoi1022/ezphot/data/scidata/7DT/7DT_C361K_HIGH_1x1/T10058/7DT14/m500/7DT14_20251102_015500_T10058_m500_1x1_100.0s_0002.fits')

# %%



#%%
from pathlib import Path
from typing import List, Tuple, Optional
from concurrent.futures import ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor

import numpy as np
from astropy.io import ascii
import astropy
from shapely.geometry import box

from ezphot.skycatalog import SkyCatalog
from ezphot.helper import Helper

#%%

class SkyCatalogUtility:
    """
    Utility class for loading, filtering, and selecting sources from sky catalogs.    
    """
    
    def __init__(self,
                 catalog_archive_path: Optional[str] = None):
        """
        Parameters
        ----------
        catalog_archive_path : Optional[str], optional
            Path to the catalog archive summary file. If not provided, the path will be set to the catalog_dir in the config.
        """
        self.helper = Helper()
        self.catalog_archive_path = Path(catalog_archive_path) if catalog_archive_path else Path(self.helper.config['CATALOG_DIR']) / 'summary.ascii_fixed_width'

    def get_catalogs(self,
                     ra: float = None,
                     dec: float = None,
                     objname: str = None,
                     fov_ra: float = 1.0,
                     fov_dec: float = 1.0,
                     catalog_type: str = 'GAIAXP',
                     catalog_version: str = 'v1',
                     fraction_criteria: float = 0.05,
                     query_when_not_archived: bool = True,
                     verbose: bool = False) -> List[SkyCatalog]:
        """
        Load SkyCatalog instances that overlap with a given coordinate and FOV.
        
        Parameters
        ----------
        ra : float
            The right ascension of the target coordinate.
        dec : float
            The declination of the target coordinate.
        objname : str, optional
            The name of the target object.
        fov_ra : float, optional
            The field of view in right ascension.
        fov_dec : float, optional
            The field of view in declination.
        catalog_type : str, optional
            The type of catalog to load.
        catalog_version : str, optional
            The version of catalog to load.
        fraction_criteria : float, optional
            The fraction of the catalog that must overlap with the target coordinate and FOV.
        query_when_not_archived : bool, optional
            Whether to query the catalog when it is not archived.
        verbose : bool, optional
            Whether to print verbose output.

        Returns
        -------
        catalogs : List[SkyCatalog]
            The list of SkyCatalog instances that overlap with the target coordinate and FOV.
        """
        def make_sky_rectangle(ra_center, dec_center, fov_ra, fov_dec):
            ra_offset = (fov_ra / 2) / np.cos(np.radians(dec_center))
            dec_offset = fov_dec / 2
            return box(ra_center - ra_offset, dec_center - dec_offset,
                       ra_center + ra_offset, dec_center + dec_offset)

        # Load catalog summary
        catalog_path = self.catalog_archive_path
        summary = ascii.read(catalog_path, format='fixed_width')
        summary = summary[(summary['catalog_type'] == catalog_type) & (summary['catalog_version'] == catalog_version)]
        
        # Fast angular pre-filter
        catalogs = None
        if objname is not None:
            summary = summary[summary['objname'] == objname]
            if len(summary) > 0:
                catalogs = [self._load_catalog_worker(row) for row in summary]
                return catalogs
            else:
                self.helper.print(f"No catalogs found for {objname}", verbose)

        if ra is not None and dec is not None:
            ra_arr = np.asarray(summary['ra'])
            dec_arr = np.asarray(summary['dec'])
            cos_dec0, sin_dec0 = np.cos(np.radians(dec)), np.sin(np.radians(dec))
            cos_dec, sin_dec = np.cos(np.radians(dec_arr)), np.sin(np.radians(dec_arr))
            delta_ra = np.radians(ra_arr - ra)
            angular_sep = np.degrees(np.arccos(np.clip(sin_dec0 * sin_dec + cos_dec0 * cos_dec * np.cos(delta_ra), -1, 1)))

            search_radius = min(5, 10 * np.sqrt(fov_ra ** 2 + fov_dec ** 2))
            summary_nearby = summary[angular_sep < search_radius]

            # Target polygon
            target_poly = make_sky_rectangle(ra, dec, fov_ra, fov_dec)

            # Intersection filter
            from shapely.ops import unary_union

            selected_rows = []
            selected_intersections = []

            # 1. 모든 intersecting tile의 intersection polygon 저장
            candidates = []

            for row in summary_nearby:
                tile_poly = make_sky_rectangle(
                    row['ra'], row['dec'], row['fov_ra'], row['fov_dec']
                )
                if tile_poly.intersects(target_poly):
                    inter = tile_poly.intersection(target_poly)
                    if inter.area > 0:
                        candidates.append((row, inter))

            # 2. intersection 면적 큰 순서로 정렬
            candidates.sort(key=lambda x: x[1].area, reverse=True)

            # 3. Greedy union
            union_poly = None
            for row, inter in candidates:
                if union_poly is None:
                    trial_union = inter
                else:
                    trial_union = union_poly.union(inter)

                old_cov = 0.0 if union_poly is None else union_poly.area / target_poly.area
                new_cov = trial_union.area / target_poly.area

                # coverage가 증가하면 채택
                if new_cov > old_cov:
                    selected_rows.append(row)
                    union_poly = trial_union

                if new_cov >= 0.999:  # 거의 full cover
                    break

            final_coverage = 0.0 if union_poly is None else union_poly.area / target_poly.area
            self.helper.print(
                f"Selected {len(selected_rows)} tiles, coverage = {final_coverage:.3f}",
                verbose
            )
            
            catalogs = [self._load_catalog_worker(row) for row in selected_rows]
        else:
            raise RuntimeError('Objname or Coordinate(Ra and dec) must be provided')
                
        if query_when_not_archived:
            if len(catalogs) == 0:
                skycatalog = SkyCatalog(objname = objname, ra = ra, dec = dec, fov_ra = fov_ra, fov_dec = fov_dec, catalog_type = catalog_type, overlapped_fraction = fraction_criteria, verbose = verbose)
                catalogs = [skycatalog]

        return catalogs

    def select_reference_sources(self,
                                 catalog: SkyCatalog,
                                 mag_lower: Optional[float] = None,
                                 mag_upper: Optional[float] = None,
                                 **kwargs) -> Tuple['astropy.table.Table', list]:
        """
        Select reference sources from a given SkyCatalog based on predefined cuts.
        
        Parameters
        ----------
        catalog : SkyCatalog
            The SkyCatalog instance to select reference sources from.
        mag_lower : Optional[float], optional
            The lower magnitude bound for filtering reference sources.
        mag_upper : Optional[float], optional
            The upper magnitude bound for filtering reference sources.
        **kwargs : dict
            Additional keyword arguments to filter reference sources.

        Returns
        -------
        ref_sources : astropy.table.Table
            The selected reference sources.
        applied_kwargs : list
            The applied keyword arguments.
        """
        if not catalog.data:
            raise RuntimeError(f'No catalog data found for {catalog.objname}')

        # Define default cutlines
        cutlines = {
            'APASS': dict(e_V_mag=[0.001, 0.2], V_mag=[mag_lower, mag_upper]),
            'GAIA': dict(V_flag=[0, 1], V_mag=[mag_lower, mag_upper]),
            'GAIAXP': {"pmra": [-20, 20], "pmdec": [-20, 20], "bp-rp": [-2, 2], "g_mean": [mag_lower, mag_upper]},
            'PS1': {"gFlags": [0, 10], "g_mag": [mag_lower, mag_upper]},
            'SMSS': {"ngood": [5, 999], "class_star": [0.3, 1.0], "g_mag": [mag_lower, mag_upper]},
            'GAIAXP_CORR_LAMOST': {"pmra": [-20, 20], "pmdec": [-20, 20], "bp-rp": [-2, 2], "g_mean": [mag_lower, mag_upper]},
        }

        if catalog.catalog_type not in cutlines:
            raise ValueError(f"Invalid catalog type: {catalog.catalog_type}")

        cutline = {**cutlines[catalog.catalog_type], **kwargs}

        ref_sources = catalog.data
        applied_kwargs = []
        for key, value in cutline.items():
            if key in ref_sources.colnames:
                applied_kwargs.append({key: [value]})
                ref_sources = ref_sources[(ref_sources[key] >= value[0]) & (ref_sources[key] <= value[1])]

        return ref_sources, applied_kwargs

    @staticmethod
    def _load_catalog_worker(row):
        """Worker to load a SkyCatalog from a summary row."""
        return SkyCatalog(objname=row['objname'],
                          catalog_type=row['catalog_type'],
                          fov_ra=row['fov_ra'],
                          fov_dec=row['fov_dec'],
                          catalog_version=row['catalog_version'],
                          verbose = False)



import requests
from astropy.table import Table
from io import StringIO
import os
import time

# ps1 image download page
ps1filename = "https://ps1images.stsci.edu/cgi-bin/ps1filenames.py"
fitscut = "https://ps1images.stsci.edu/cgi-bin/fitscut.cgi"

#   Function
# downloading images from ps1
# tra, tdec:center coordinate of the image(in degrees)
# size: image pixel size, prefered less than 6000 pix
# filters: needed filters (grizy)
def getimages(tra, tdec, size=240, filters="grizy", format="fits", imagetypes="stack"):
    if format not in ("jpg", "png", "fits"):
        raise ValueError("format must be one of jpg, png, fits")
    if not isinstance(imagetypes, str):
        imagetypes = ",".join(imagetypes)
    cbuf = StringIO()
    cbuf.write('\n'.join(["{} {}".format(ra, dec) for (ra, dec) in zip(tra, tdec)]))
    cbuf.seek(0)
    r = requests.post(ps1filename, data=dict(filters=filters, type=imagetypes),
                      files=dict(file=cbuf))
    r.raise_for_status()
    tab = Table.read(r.text, format="ascii")

    urlbase = "{}?size={}&format={}".format(fitscut, size, format)
    tab["url"] = ["{}&ra={}&dec={}&red={}".format(urlbase, ra, dec, filename)
                  for (filename, ra, dec) in zip(tab["filename"], tab["ra"], tab["dec"])]
    return tab

def download_images_for_pointings(tab, path_slice, n_retry=3):
    """
    Download images for a list of RA/Dec pointings.
    
    Parameters:
    - ra_dec_pairs (list of tuples): List of (RA, Dec) pairs for image downloading.
    - filters (str): Filter to use for image query (e.g., 'g', 'r', 'i', 'z', 'y').
    - path_slice (str): Path to save downloaded images.
    - n_retry (int): Number of retries for failed downloads.
    
    Returns:
    - image_files (list): List of paths to downloaded image files.
    """
    image_files = []
    image_counter = 1
    for rr, row in enumerate(tab):
        ra = row['ra']
        dec = row['dec']
        projcell = row['projcell']
        subcell = row['subcell']
        filte = row['filter']
        t0 = time.time()
        fname = "t{:08.4f}{:+07.4f}.{}.fits".format(ra, dec, filte)
        file_path = os.path.join(path_slice, fname)

        if os.path.exists(file_path) and os.path.getsize(file_path) > 0:
            print(f"[{rr:0>2}] {time.time() - t0:.1f} s: Image already exists and is complete: {fname}", end='-->')
        else:
            if os.path.exists(file_path):
                print(f"[{rr:0>2}] {time.time() - t0:.1f} s: Incomplete image detected, re-downloading: {fname}", end='-->')
            else:
                print(f"[{rr:0>2}] {time.time() - t0:.1f} s: Downloading missing image: {fname}", end='-->')

            url = row["url"]

            n_try = 0
            query_success = False
            while (n_try < n_retry) and (query_success == False):
                try:
                    response = requests.get(url, timeout=(5, 10))
                    with open(file_path, "wb") as file:
                        file.write(response.content)
                    query_success = True
                    n_try = 0
                except Exception as e:
                    print(f"Error Occured:")
                    print(e)
                    print(f"Retry...")
                    n_try += 1

        image_files.append(file_path)
        print(f"Done!")
        image_counter += 1
    return image_files



#%%
import inspect
import os
from pathlib import Path
from multiprocessing import Pool

import numpy as np
import astropy.units as u
from astropy.coordinates import Longitude, Latitude, Angle, SkyCoord
from astropy.wcs import WCS
from astroquery.hips2fits import hips2fits
from astroquery.mocserver import MOCServer
from matplotlib import cm
#%%
hips2fits.timeout = 300
class HIPS2FITS:
    """
    A class to handle image queries.
    """

    def __init__(self, catalog_key: str = None):
        """
        Initializes the ImageQuerier with the path to the image.

        :param image_path: Path to the image file.
        """
        if catalog_key is not None:
            if catalog_key not in self.catalog_ids.keys():
                raise ValueError(f"Catalog Key '{catalog_key}' is not recognized. Available keys: {list(self.catalog_ids.keys())}")
        self.current_catalog_key = catalog_key
    
    @property
    def config(self):
        class Configuration:
            """Handles configuration for Vizier queries."""

            @property
            def projection(self):
                return 'TAN'
            
            @projection.setter
            def projection(self, value):
                self.projection = value
            
            @property
            def coordsys(self):
                return 'icrs'
        
            @coordsys.setter
            def coordsys(self, value):
                self.coordsys = value
            
            @property
            def format(self):
                return 'fits'
            
            @format.setter
            def format(self, value):
                self.format = value
                
            @property
            def stretch(self):
                return 'linear'
            
            @stretch.setter
            def stretch(self, value):
                self.stretch = value
                
            @property
            def cmap(self):
                return 'Greys_r'

            @cmap.setter
            def cmap(self, value):
                self.cmap = value
                
            @property
            def min_cut(self):
                return 0.5
            
            @min_cut.setter
            def min_cut(self, value):
                self.min_cut = value
                
            @property
            def max_cut(self):
                return 99.5
            
            @max_cut.setter
            def max_cut(self, value):
                self.max_cut = value
            
            def __repr__(self):
                return (f"========HIPS2FITS Configuration========\n"
                        f"  projection      = {self.projection}\n"
                        f"  coordsys        = {self.coordsys}\n"
                        f"  format          = {self.format}\n"
                        f"  stretch         = {self.stretch}\n"
                        f"  cmap            = {self.cmap}\n"
                        f"  min_cut         = {self.min_cut}\n"
                        f"  max_cut         = {self.max_cut}\n"
                        f"====================================")
        
        return Configuration()
    
    @property
    def catalog_ids(self):
        catalog_ids = dict()
        # SkyMapper DR4 
        catalog_ids['SkyMapper/SMSS4/g'] = 'CDS/P/Skymapper/DR4/g'
        catalog_ids['SkyMapper/SMSS4/r'] = 'CDS/P/Skymapper/DR4/r'
        catalog_ids['SkyMapper/SMSS4/i'] = 'CDS/P/Skymapper/DR4/i'
        
        # Skymapper DR1
        catalog_ids['SkyMapper/SMSS1/u'] = 'CDS/P/Skymapper-U'
        catalog_ids['SkyMapper/SMSS1/g'] = 'CDS/P/Skymapper-G'
        catalog_ids['SkyMapper/SMSS1/v'] = 'CDS/P/Skymapper-V'
        catalog_ids['SkyMapper/SMSS1/r'] = 'CDS/P/Skymapper-R'
        catalog_ids['SkyMapper/SMSS1/i'] = 'CDS/P/Skymapper-I'
        catalog_ids['SkyMapper/SMSS1/z'] = 'CDS/P/Skymapper-Z'

        # Pan-STARRS DR1 
        catalog_ids['PanSTARRS/PS1/g'] = "CDS/P/PanSTARRS/DR1/g"
        catalog_ids['PanSTARRS/PS1/r'] = "CDS/P/PanSTARRS/DR1/r"
        catalog_ids['PanSTARRS/PS1/i'] = "CDS/P/PanSTARRS/DR1/i"
        catalog_ids['PanSTARRS/PS1/z'] = "CDS/P/PanSTARRS/DR1/z"
        catalog_ids['PanSTARRS/PS1/y'] = "CDS/P/PanSTARRS/DR1/y"
        
        # SDSS DR9 
        catalog_ids['SDSS/SDSS9/u'] = "CDS/P/SDSS9/u"
        catalog_ids['SDSS/SDSS9/g'] = "CDS/P/SDSS9/g"
        catalog_ids['SDSS/SDSS9/r'] = "CDS/P/SDSS9/r"
        catalog_ids['SDSS/SDSS9/i'] = "CDS/P/SDSS9/i"
        catalog_ids['SDSS/SDSS9/z'] = "CDS/P/SDSS9/z"
        
        # DESI Legacy Imaging Survey
        catalog_ids['DESI/DESI/g'] = "CDS/P/DESI-Legacy-Surveys/DR10/g"
        catalog_ids['DESI/DESI/r'] = "CDS/P/DESI-Legacy-Surveys/DR10/r"
        catalog_ids['DESI/DESI/i'] = "CDS/P/DESI-Legacy-Surveys/DR10/i"
        catalog_ids['DESI/DESI/z'] = "CDS/P/DESI-Legacy-Surveys/DR10/z"
        
        # DSS 
        catalog_ids['DSS/DSS2/b'] = "CDS/P/DSS2/blue"
        catalog_ids['DSS/DSS2/r'] = "CDS/P/DSS2/red"
        # catalog_ids['DSSDSS2/nir'] = "CDS/P/DSS2/NIR"
        
        # ZTF
        catalog_ids['ZTF/ZTF7/g'] = "CDS/P/ZTF/DR7/g"
        catalog_ids['ZTF/ZTF7/r'] = "CDS/P/ZTF/DR7/r"
        catalog_ids['ZTF/ZTF7/i'] = "CDS/P/ZTF/DR7/i"
        
        # DECAM
        catalog_ids['DECALS/DEC5/g'] = "CDS/P/DECaLS/DR5/g"
        catalog_ids['DECALS/DEC5/r'] = "CDS/P/DECaLS/DR5/r"
        
        # DES
        catalog_ids['DES/DES2/g'] = "CDS/P/DES-DR2/g"
        catalog_ids['DES/DES2/r'] = "CDS/P/DES-DR2/r"
        catalog_ids['DES/DES2/i'] = "CDS/P/DES-DR2/i"
        catalog_ids['DES/DES2/z'] = "CDS/P/DES-DR2/z"
        catalog_ids['DES/DES2/Y'] = "CDS/P/DES-DR2/Y"
        
        return catalog_ids

    def _change_catalog(self, catalog_key):
        if catalog_key in self.catalog_ids.keys():
            self.current_catalog_key = catalog_key
            print(self.__repr__())
        else:
            raise ValueError(f"Catalog Key '{catalog_key}' is not recognized.")

    def _show_available_catalogs(self):
        print("Current catalog: ", self.current_catalog_key)
        print("Available catalogs\n==================")
        for catalog_name, catalog_id in self.catalog_ids.items():
            print(f"{catalog_name}: {catalog_id}")
        return list(self.catalog_ids.keys())
    
    def _check_catalog_coverages(self, 
                                 ra: float,
                                 dec: float, 
                                 radius_deg: float = 0.1, 
                                 verbose: bool = True,
                                 search_all: bool = False) -> dict:
        from regions import CircleSkyRegion

        coord = SkyCoord(ra=ra * u.deg, dec=dec * u.deg)
        region = CircleSkyRegion(center=coord, radius=radius_deg * u.deg)
        results = {}

        if not search_all:
            try:
                catalog_key = self.current_catalog_key
                hips_id = self.catalog_ids[catalog_key]
                query = MOCServer.query_region(
                    region=region,
                    criteria=f"ID={hips_id}",
                    intersect="overlaps",
                    max_rec=1
                )
                has_coverage = len(query) > 0
            except Exception as e:
                has_coverage = False
                if verbose:
                    print(f"[ERROR] Failed to check {catalog_key}: {e}")
            results[catalog_key] = has_coverage
            if verbose:
                print(f"[{catalog_key}] {'✓' if has_coverage else '✗'}")
        else:
            for catalog_key, hips_id in self.catalog_ids.items():
                try:
                    query = MOCServer.query_region(
                        region=region,
                        criteria=f"ID={hips_id}",
                        intersect="overlaps",
                        max_rec=1
                    )
                    has_coverage = len(query) > 0
                except Exception as e:
                    has_coverage = False
                    if verbose:
                        print(f"[ERROR] Failed to check {catalog_key}: {e}")
                results[catalog_key] = has_coverage
                if verbose:
                    print(f"[{catalog_key}] {'✓' if has_coverage else '✗'}")

        return results

    def _query(self,
               wcs: WCS = None, # If wcs inputted, overrides ra, dec, fov, rotation_angle
               width: int = 2000,
               height: int = 2000,
               ra: float = 0.0,
               dec: float = 0.0,
               fov: float = 5.0,
               rotation_angle: float = 0.0,
               save_path: str = None,
               verbose: bool = False,
               ):
        
        if save_path is None:
            save_path = os.path.join(os.getcwd(), f"hips2fits_{self.current_catalog_key}_{ra}_{dec}.fits")
            if verbose:
                print(f"Default save path: {save_path}")
        
        if wcs is not None:
            # If WCS is provided, use it to query the image
            result = hips2fits.query_with_wcs(
                hips=self.catalog_ids[self.current_catalog_key],
                wcs=wcs,
                format=self.config.format,
                min_cut=self.config.min_cut,
                max_cut=self.config.max_cut,
                stretch=self.config.stretch,
                cmap=cm.get_cmap(self.config.cmap),
                verbose=verbose,
            )
        else:
            result = hips2fits.query(
                hips=self.catalog_ids[self.current_catalog_key],
                width=width,
                height=height,
                projection = self.config.projection,
                fov=Angle(fov * u.deg),
                ra=Longitude(ra * u.deg),
                dec=Latitude(dec * u.deg),
                coordsys=self.config.coordsys,
                rotation_angle=Angle(rotation_angle * u.deg),
                format=self.config.format,
                min_cut=self.config.min_cut,
                max_cut=self.config.max_cut,
                stretch=self.config.stretch,
                cmap=cm.get_cmap(self.config.cmap),
                verbose=verbose,
            )
            
        if verbose:
            print(f"Saved: {save_path}")
        result[0].writeto(save_path, overwrite=True)
        return save_path
            

class ImageQuerier(HIPS2FITS):
    """
    A class to handle image queries using HiPS2FITS.
    Inherits from HIPS2FITS to utilize its methods and properties.
    """

    def __init__(self, catalog_key: str = None):
        """
        Initializes the ImageQuerier with the specified catalog key.

        :param catalog_key: Key for the catalog to query.
        """
        super().__init__(catalog_key=catalog_key)

    def __repr__(self):
        return f"ImageQuerier(catalog={self.current_catalog_key})\n{self.config}\n For help, use 'help(self)' or `self.help()`. \nPlease check the HIPS2FITS documentation for more information. \n(https://astroquery.readthedocs.io/en/latest/hips2fits/hips2fits.html)"

    def help(self):
        # Get all public methods from the class, excluding `help`
        methods = [
            (name, obj)
            for name, obj in inspect.getmembers(self.__class__, inspect.isfunction)
            if not name.startswith("_") and name != "help"
        ]

        # Build plain text list with parameters
        lines = []
        for name, func in methods:
            sig = inspect.signature(func)
            params = [str(p) for p in sig.parameters.values() if p.name != "self"]
            sig_str = f"({', '.join(params)})" if params else "()"
            lines.append(f"- {name}{sig_str}")

        # Final plain text output
        help_text = ""
        print(f"Help for {self.__class__.__name__}\n{help_text}\n\nPublic methods:\n" + "\n".join(lines))

    def query(self,
              width: int,
              height: int,
              ra: float,
              dec: float,
              pixelscale: float,
              save_path: str = None,
              rotation_angle: float = 0.0,
              verbose: bool = True,
              n_processes: int = 4):
        """
        Run HiPS2FITS queries with split tiles using multiprocessing.
        
        Parameters
        ----------
        width : int
            Width of the image in pixels
        height : int
            Height of the image in pixels
        ra : float
            Right Ascension in degrees
        dec : float
            Declination in degrees
        pixelscale : float
            Pixel scale in degrees per pixel
        save_path : str
            Path to save the image
        rotation_angle : float
            Rotation angle in degrees
        verbose : bool
            If True, prints verbose output
        n_processes : int
            Number of processes to use for multiprocessing

        Returns
        -------
        list of str : Paths to the saved images
        """
        if self.current_catalog_key is None:
            raise ValueError("No catalog key provided. Please set the catalog key using the change_catalog method.")
        if self.check_coverage(ra, dec,search_all = False)[self.current_catalog_key] is False:
            raise ValueError(f"No coverage found for {self.current_catalog_key} at {ra}, {dec}")
        observatory, telname, filter_ = self.current_catalog_key.split('/')

        fov = max(width, height) * pixelscale / 3600  # Convert pixel scale to degrees
        tile_params = self._split_query_regions(
            width=width,
            height=height,
            ra=ra,
            dec=dec,
            fov=fov,
            rotation_angle=rotation_angle,
            verbose=verbose
        )
        if save_path is None:
            output_path = os.path.join(Path.home(), f"hips2fits_{observatory}_{telname}_{ra}_{dec}.fits")
        else:
            output_path = str(save_path)
        if verbose:
            print(f"[QUERY] Dispatching {len(tile_params)} tiles with {n_processes} processes")

        with Pool(processes=n_processes) as pool:
            tasks = [(param, output_path, verbose) for param in tile_params]
            results = pool.map(self._query_tile_worker, tasks)
        
        return results
    
    def check_coverage(self,
                       ra: float,
                       dec: float,
                       radius_deg: float = 1.0,
                       search_all: bool = False,
                       verbose: bool = True):
        """
        Check if the current HiPS catalog has coverage at the given RA/Dec.
        
        Parameters
        ----------
        ra : float
            Right Ascension in degrees
        dec : float
            Declination in degrees
        radius_deg : float
            Search radius in degrees
        search_all : bool
            If True, searches all catalogs
        verbose : bool
            If True, prints whether coverage exists

        Returns
        -------
        coverage_dict : dict
            Dictionary of catalog_key -> bool (True if covered).
        """
        return self._check_catalog_coverages(ra=ra, dec=dec, radius_deg=radius_deg, verbose=verbose, search_all=search_all)
    
    def change_catalog(self, catalog_key):
        """
        Change the current catalog to query.
        
        Parameters
        ----------
        catalog_key : str
            The catalog key to change to.
        
        Returns
        -------
        None
        """
        self._change_catalog(catalog_key)
    
    def show_available_catalogs(self):
        """
        Display available catalogs.
        
        Parameters
        ----------
        None
                
        Returns
        -------
        list of catalog keys: list
            List of available catalogs.
        """
        return self._show_available_catalogs()
    
    def _split_query_regions(self,
                             width: int,
                             height: int,
                             ra: float,
                             dec: float,
                             fov: float,
                             rotation_angle: float = 0.0,
                             max_pixels: int = 45000000,
                             margin_fraction: float = 0.1,
                             verbose: bool = True):
        """
        Split a large query into tiles with unique RA/Dec centers that cover the full region.

        Returns
        -------
        list of dict : Each dict contains query parameters for a tile
        """
        total_pixels = width * height

        if total_pixels <= max_pixels:
            width_with_margin = int(np.ceil(width * (1 + margin_fraction)))
            height_with_margin = int(np.ceil(height * (1 + margin_fraction)))
            fov_with_margin = fov * (1 + margin_fraction)
            return [{
                'width': width_with_margin,
                'height': height_with_margin,
                'ra': ra,
                'dec': dec,
                'fov': fov_with_margin,
                'rotation_angle': rotation_angle,
                'tile_id': '0_0'
            }]

        # Determine number of splits needed per axis
        n_splits = int(np.ceil(np.sqrt(total_pixels / max_pixels)))
        tile_width = width // n_splits
        tile_height = height // n_splits
        base_tile_fov = fov / n_splits  # Original FoV per tile

        # Apply margin to width/height and FoV
        tile_width_with_margin = int(np.ceil(tile_width * (1 + margin_fraction)))
        tile_height_with_margin = int(np.ceil(tile_height * (1 + margin_fraction)))
        tile_fov_with_margin = base_tile_fov * (1 + margin_fraction)

        # Pixel scale (deg/pixel) stays fixed
        pixscale_deg = fov / width
        dec_rad = np.deg2rad(dec)

        if verbose:
            print(f"[SPLIT] {width}x{height} >>> {n_splits}x{n_splits} tiles "
                f"({tile_width}x{tile_height} px), margin = {margin_fraction:.1%}")

        tile_params = []

        for i in range(n_splits):
            for j in range(n_splits):
                # Offset in sky from the center (no margin added here)
                delta_ra_deg = ((j + 0.5) - n_splits / 2) * tile_width * pixscale_deg / np.cos(dec_rad)
                delta_dec_deg = ((n_splits / 2) - (i + 0.5)) * tile_height * pixscale_deg

                center_ra = ra + delta_ra_deg
                center_dec = dec + delta_dec_deg

                tile_params.append({
                    'width': tile_width_with_margin,
                    'height': tile_height_with_margin,
                    'ra': center_ra,
                    'dec': center_dec,
                    'fov': tile_fov_with_margin,
                    'rotation_angle': rotation_angle,
                    'tile_id': f"{i}_{j}"
                })

        return tile_params

    def _query_tile_worker(self, kwargs):
        """
        Standalone worker function for multiprocessing.
        This function must be top-level (not class method) for pickling.
        """
        tile_param, save_path, verbose = kwargs
        save_path = Path(save_path)

        save_path = save_path.with_suffix(f".{tile_param['tile_id']}.fits")

        return self._query(
            wcs=None,
            width=tile_param['width'],
            height=tile_param['height'],
            ra=tile_param['ra'],
            dec=tile_param['dec'],
            fov=tile_param['fov'],
            rotation_angle=tile_param['rotation_angle'],
            verbose=verbose,
            save_path=save_path
        )

